@@ -25,16 +25,25 @@ import (
 	"encoding/json"
 	"fmt"
 	"net/url"
+	"sort"
 	"strings"
+
+	"github.com/golang-jwt/jwt/v4"
 )
 
 const (
-	// Version that must be sent in a "hello" message.
-	HelloVersion = "1.0"
+	// Version 1.0 validates auth params against the Nextcloud instance.
+	HelloVersionV1 = "1.0"
+
+	// Version 2.0 validates auth params encoded as JWT.
+	HelloVersionV2 = "2.0"
 )
 
 // ClientMessage is a message that is sent from a client to the server.
 type ClientMessage struct {
+	json.Marshaler
+	json.Unmarshaler
+
 	// The unique request id (optional).
 	Id string `json:"id,omitempty"`
 
@@ -53,6 +62,8 @@ type ClientMessage struct {
 	Control *ControlClientMessage `json:"control,omitempty"`
 
 	Internal *InternalClientMessage `json:"internal,omitempty"`
+
+	TransientData *TransientDataClientMessage `json:"transient,omitempty"`
 }
 
 func (m *ClientMessage) CheckValid() error {
@@ -91,6 +102,12 @@ func (m *ClientMessage) CheckValid() error {
 		} else if err := m.Internal.CheckValid(); err != nil {
 			return err
 		}
+	case "transient":
+		if m.TransientData == nil {
+			return fmt.Errorf("transient missing")
+		} else if err := m.TransientData.CheckValid(); err != nil {
+			return err
+		}
 	}
 	return nil
 }
@@ -121,12 +138,17 @@ func (m *ClientMessage) NewWrappedErrorServerMessage(e error) *ServerMessage {
 
 // ServerMessage is a message that is sent from the server to a client.
 type ServerMessage struct {
+	json.Marshaler
+	json.Unmarshaler
+
 	Id string `json:"id,omitempty"`
 
 	Type string `json:"type"`
 
 	Error *Error `json:"error,omitempty"`
 
+	Welcome *WelcomeServerMessage `json:"welcome,omitempty"`
+
 	Hello *HelloServerMessage `json:"hello,omitempty"`
 
 	Bye *ByeServerMessage `json:"bye,omitempty"`
@@ -138,6 +160,8 @@ type ServerMessage struct {
 	Control *ControlServerMessage `json:"control,omitempty"`
 
 	Event *EventServerMessage `json:"event,omitempty"`
+
+	TransientData *TransientDataServerMessage `json:"transient,omitempty"`
 }
 
 func (r *ServerMessage) CloseAfterSend(session Session) bool {
@@ -217,6 +241,54 @@ func (e *Error) Error() string {
 	return e.Message
 }
 
+type WelcomeServerMessage struct {
+	Version  string   `json:"version"`
+	Features []string `json:"features,omitempty"`
+	Country  string   `json:"country,omitempty"`
+}
+
+func NewWelcomeServerMessage(version string, feature ...string) *WelcomeServerMessage {
+	message := &WelcomeServerMessage{
+		Version:  version,
+		Features: feature,
+	}
+	if len(feature) > 0 {
+		sort.Strings(message.Features)
+	}
+	return message
+}
+
+func (m *WelcomeServerMessage) AddFeature(feature ...string) {
+	newFeatures := make([]string, len(m.Features))
+	copy(newFeatures, m.Features)
+	for _, feat := range feature {
+		found := false
+		for _, f := range newFeatures {
+			if f == feat {
+				found = true
+				break
+			}
+		}
+		if !found {
+			newFeatures = append(newFeatures, feat)
+		}
+	}
+	sort.Strings(newFeatures)
+	m.Features = newFeatures
+}
+
+func (m *WelcomeServerMessage) RemoveFeature(feature ...string) {
+	newFeatures := make([]string, len(m.Features))
+	copy(newFeatures, m.Features)
+	for _, feat := range feature {
+		idx := sort.SearchStrings(newFeatures, feat)
+		if idx < len(newFeatures) && newFeatures[idx] == feat {
+			newFeatures = append(newFeatures[:idx], newFeatures[idx+1:]...)
+		}
+	}
+	m.Features = newFeatures
+}
+
 const (
 	HelloClientTypeClient   = "client"
 	HelloClientTypeInternal = "internal"
@@ -258,6 +330,23 @@ func (p *ClientTypeInternalAuthParams) CheckValid() error {
 	return nil
 }
 
+type HelloV2AuthParams struct {
+	Token string `json:"token"`
+}
+
+func (p *HelloV2AuthParams) CheckValid() error {
+	if p.Token == "" {
+		return fmt.Errorf("token missing")
+	}
+	return nil
+}
+
+type HelloV2TokenClaims struct {
+	jwt.RegisteredClaims
+
+	UserData *json.RawMessage `json:"userdata,omitempty"`
+}
+
 type HelloClientMessageAuth struct {
 	// The client type that is connecting. Leave empty to use the default
 	// "HelloClientTypeClient"
@@ -269,6 +358,7 @@ type HelloClientMessageAuth struct {
 	parsedUrl *url.URL
 
 	internalParams ClientTypeInternalAuthParams
+	helloV2Params  HelloV2AuthParams
 }
 
 // Type "hello"
@@ -285,8 +375,8 @@ type HelloClientMessage struct {
 }
 
 func (m *HelloClientMessage) CheckValid() error {
-	if m.Version != HelloVersion {
-		return fmt.Errorf("unsupported hello version: %s", m.Version)
+	if m.Version != HelloVersionV1 && m.Version != HelloVersionV2 {
+		return InvalidHelloVersion
 	}
 	if m.ResumeId == "" {
 		if m.Auth.Params == nil || len(*m.Auth.Params) == 0 {
@@ -308,6 +398,17 @@ func (m *HelloClientMessage) CheckValid() error {
 
 				m.Auth.parsedUrl = u
 			}
+
+			switch m.Version {
+			case HelloVersionV1:
+				// No additional validation necessary.
+			case HelloVersionV2:
+				if err := json.Unmarshal(*m.Auth.Params, &m.Auth.helloV2Params); err != nil {
+					return err
+				} else if err := m.Auth.helloV2Params.CheckValid(); err != nil {
+					return err
+				}
+			}
 		case HelloClientTypeInternal:
 			if err := json.Unmarshal(*m.Auth.Params, &m.Auth.internalParams); err != nil {
 				return err
@@ -322,37 +423,61 @@ func (m *HelloClientMessage) CheckValid() error {
 }
 
 const (
-	// Features for all clients.
+	// Features to send to all clients.
 	ServerFeatureMcu                   = "mcu"
 	ServerFeatureSimulcast             = "simulcast"
+	ServerFeatureUpdateSdp             = "update-sdp"
 	ServerFeatureAudioVideoPermissions = "audio-video-permissions"
+	ServerFeatureTransientData         = "transient-data"
+	ServerFeatureInCallAll             = "incall-all"
+	ServerFeatureWelcome               = "welcome"
+	ServerFeatureHelloV2               = "hello-v2"
+	ServerFeatureSwitchTo              = "switchto"
 
-	// Features for internal clients only.
+	// Features to send to internal clients only.
 	ServerFeatureInternalVirtualSessions = "virtual-sessions"
+
+	// Possible client features from the "hello" request.
+	ClientFeatureInternalInCall = "internal-incall"
 )
 
 var (
-	DefaultFeatures []string = []string{
+	DefaultFeatures = []string{
 		ServerFeatureAudioVideoPermissions,
+		ServerFeatureTransientData,
+		ServerFeatureInCallAll,
+		ServerFeatureWelcome,
+		ServerFeatureHelloV2,
+		ServerFeatureSwitchTo,
 	}
-	DefaultFeaturesInternal []string = []string{
+	DefaultFeaturesInternal = []string{
 		ServerFeatureInternalVirtualSessions,
+		ServerFeatureTransientData,
+		ServerFeatureInCallAll,
+		ServerFeatureWelcome,
+		ServerFeatureHelloV2,
+		ServerFeatureSwitchTo,
+	}
+	DefaultWelcomeFeatures = []string{
+		ServerFeatureAudioVideoPermissions,
+		ServerFeatureInternalVirtualSessions,
+		ServerFeatureTransientData,
+		ServerFeatureInCallAll,
+		ServerFeatureWelcome,
+		ServerFeatureHelloV2,
+		ServerFeatureSwitchTo,
 	}
 )
 
-type HelloServerMessageServer struct {
-	Version  string   `json:"version"`
-	Features []string `json:"features,omitempty"`
-	Country  string   `json:"country,omitempty"`
-}
-
 type HelloServerMessage struct {
 	Version string `json:"version"`
 
-	SessionId string                    `json:"sessionid"`
-	ResumeId  string                    `json:"resumeid"`
-	UserId    string                    `json:"userid"`
-	Server    *HelloServerMessageServer `json:"server,omitempty"`
+	SessionId string `json:"sessionid"`
+	ResumeId  string `json:"resumeid"`
+	UserId    string `json:"userid"`
+
+	// TODO: Remove once all clients have switched to the "welcome" message.
+	Server *WelcomeServerMessage `json:"server,omitempty"`
 }
 
 // Type "bye"
@@ -467,10 +592,7 @@ type ControlClientMessage struct {
 }
 
 func (m *ControlClientMessage) CheckValid() error {
-	if err := m.MessageClientMessage.CheckValid(); err != nil {
-		return err
-	}
-	return nil
+	return m.MessageClientMessage.CheckValid()
 }
 
 type ControlServerMessage struct {
@@ -509,28 +631,24 @@ type AddSessionInternalClientMessage struct {
 	UserId string           `json:"userid,omitempty"`
 	User   *json.RawMessage `json:"user,omitempty"`
 	Flags  uint32           `json:"flags,omitempty"`
+	InCall *int             `json:"incall,omitempty"`
 
 	Options *AddSessionOptions `json:"options,omitempty"`
 }
 
 func (m *AddSessionInternalClientMessage) CheckValid() error {
-	if err := m.CommonSessionInternalClientMessage.CheckValid(); err != nil {
-		return err
-	}
-	return nil
+	return m.CommonSessionInternalClientMessage.CheckValid()
 }
 
 type UpdateSessionInternalClientMessage struct {
 	CommonSessionInternalClientMessage
 
-	Flags *uint32 `json:"flags,omitempty"`
+	Flags  *uint32 `json:"flags,omitempty"`
+	InCall *int    `json:"incall,omitempty"`
 }
 
 func (m *UpdateSessionInternalClientMessage) CheckValid() error {
-	if err := m.CommonSessionInternalClientMessage.CheckValid(); err != nil {
-		return err
-	}
-	return nil
+	return m.CommonSessionInternalClientMessage.CheckValid()
 }
 
 type RemoveSessionInternalClientMessage struct {
@@ -540,9 +658,14 @@ type RemoveSessionInternalClientMessage struct {
 }
 
 func (m *RemoveSessionInternalClientMessage) CheckValid() error {
-	if err := m.CommonSessionInternalClientMessage.CheckValid(); err != nil {
-		return err
-	}
+	return m.CommonSessionInternalClientMessage.CheckValid()
+}
+
+type InCallInternalClientMessage struct {
+	InCall int `json:"incall"`
+}
+
+func (m *InCallInternalClientMessage) CheckValid() error {
 	return nil
 }
 
@@ -554,6 +677,8 @@ type InternalClientMessage struct {
 	UpdateSession *UpdateSessionInternalClientMessage `json:"updatesession,omitempty"`
 
 	RemoveSession *RemoveSessionInternalClientMessage `json:"removesession,omitempty"`
+
+	InCall *InCallInternalClientMessage `json:"incall,omitempty"`
 }
 
 func (m *InternalClientMessage) CheckValid() error {
@@ -576,6 +701,12 @@ func (m *InternalClientMessage) CheckValid() error {
 		} else if err := m.RemoveSession.CheckValid(); err != nil {
 			return err
 		}
+	case "incall":
+		if m.InCall == nil {
+			return fmt.Errorf("incall missing")
+		} else if err := m.InCall.CheckValid(); err != nil {
+			return err
+		}
 	}
 	return nil
 }
@@ -589,6 +720,16 @@ type RoomEventServerMessage struct {
 	InCall  *json.RawMessage         `json:"incall,omitempty"`
 	Changed []map[string]interface{} `json:"changed,omitempty"`
 	Users   []map[string]interface{} `json:"users,omitempty"`
+
+	All bool `json:"all,omitempty"`
+}
+
+func (m *RoomEventServerMessage) String() string {
+	data, err := json.Marshal(m)
+	if err != nil {
+		return fmt.Sprintf("Could not serialize %#v: %s", m, err)
+	}
+	return string(data)
 }
 
 const (
@@ -613,14 +754,27 @@ type RoomFlagsServerMessage struct {
 	Flags     uint32 `json:"flags"`
 }
 
+type ChatComment map[string]interface{}
+
+type RoomEventMessageDataChat struct {
+	Comment *ChatComment `json:"comment,omitempty"`
+}
+
+type RoomEventMessageData struct {
+	Type string `json:"type"`
+
+	Chat *RoomEventMessageDataChat `json:"chat,omitempty"`
+}
+
 type EventServerMessage struct {
 	Target string `json:"target"`
 	Type   string `json:"type"`
 
 	// Used for target "room"
-	Join   []*EventServerMessageSessionEntry `json:"join,omitempty"`
-	Leave  []string                          `json:"leave,omitempty"`
-	Change []*EventServerMessageSessionEntry `json:"change,omitempty"`
+	Join     []*EventServerMessageSessionEntry `json:"join,omitempty"`
+	Leave    []string                          `json:"leave,omitempty"`
+	Change   []*EventServerMessageSessionEntry `json:"change,omitempty"`
+	SwitchTo *EventServerMessageSwitchTo       `json:"switchto,omitempty"`
 
 	// Used for target "roomlist" / "participants"
 	Invite    *RoomEventServerMessage          `json:"invite,omitempty"`
@@ -632,6 +786,14 @@ type EventServerMessage struct {
 	Message *RoomEventMessage `json:"message,omitempty"`
 }
 
+func (m *EventServerMessage) String() string {
+	data, err := json.Marshal(m)
+	if err != nil {
+		return fmt.Sprintf("Could not serialize %#v: %s", m, err)
+	}
+	return string(data)
+}
+
 type EventServerMessageSessionEntry struct {
 	SessionId     string           `json:"sessionid"`
 	UserId        string           `json:"userid"`
@@ -639,6 +801,20 @@ type EventServerMessageSessionEntry struct {
 	RoomSessionId string           `json:"roomsessionid,omitempty"`
 }
 
+func (e *EventServerMessageSessionEntry) Clone() *EventServerMessageSessionEntry {
+	return &EventServerMessageSessionEntry{
+		SessionId:     e.SessionId,
+		UserId:        e.UserId,
+		User:          e.User,
+		RoomSessionId: e.RoomSessionId,
+	}
+}
+
+type EventServerMessageSwitchTo struct {
+	RoomId  string          `json:"roomid"`
+	Details json.RawMessage `json:"details,omitempty"`
+}
+
 // MCU-related types
 
 type AnswerOfferMessage struct {
@@ -647,4 +823,38 @@ type AnswerOfferMessage struct {
 	Type     string                 `json:"type"`
 	RoomType string                 `json:"roomType"`
 	Payload  map[string]interface{} `json:"payload"`
+	Sid      string                 `json:"sid,omitempty"`
+}
+
+// Type "transient"
+
+type TransientDataClientMessage struct {
+	Type string `json:"type"`
+
+	Key   string           `json:"key,omitempty"`
+	Value *json.RawMessage `json:"value,omitempty"`
+}
+
+func (m *TransientDataClientMessage) CheckValid() error {
+	switch m.Type {
+	case "set":
+		if m.Key == "" {
+			return fmt.Errorf("key missing")
+		}
+		// A "nil" value is allowed and will remove the key.
+	case "remove":
+		if m.Key == "" {
+			return fmt.Errorf("key missing")
+		}
+	}
+	return nil
+}
+
+type TransientDataServerMessage struct {
+	Type string `json:"type"`
+
+	Key      string                 `json:"key,omitempty"`
+	OldValue interface{}            `json:"oldvalue,omitempty"`
+	Value    interface{}            `json:"value,omitempty"`
+	Data     map[string]interface{} `json:"data,omitempty"`
 }
