@@ -33,8 +33,7 @@ import (
 	"time"
 	"unsafe"
 
-	"github.com/nats-io/nats.go"
-	"github.com/pion/sdp"
+	"github.com/pion/sdp/v3"
 )
 
 var (
@@ -49,9 +48,10 @@ var (
 
 type ClientSession struct {
 	roomJoinTime int64
+	inCall       uint32
 
-	running   int32
 	hub       *Hub
+	events    AsyncEvents
 	privateId string
 	publicId  string
 	data      *SessionIdData
@@ -68,10 +68,7 @@ type ClientSession struct {
 	backendUrl       string
 	parsedBackendUrl *url.URL
 
-	natsReceiver chan *nats.Msg
-	stopRun      chan bool
-	runStopped   chan bool
-	expires      time.Time
+	expires time.Time
 
 	mu sync.Mutex
 
@@ -79,9 +76,7 @@ type ClientSession struct {
 	room          unsafe.Pointer
 	roomSessionId string
 
-	userSubscription    NatsSubscription
-	sessionSubscription NatsSubscription
-	roomSubscription    NatsSubscription
+	publisherWaiters ChannelWaiters
 
 	publishers  map[string]McuPublisher
 	subscribers map[string]McuSubscriber
@@ -91,11 +86,15 @@ type ClientSession struct {
 	hasPendingParticipantsUpdate bool
 
 	virtualSessions map[*VirtualSession]bool
+
+	seenJoinedLock   sync.Mutex
+	seenJoinedEvents map[string]bool
 }
 
 func NewClientSession(hub *Hub, privateId string, publicId string, data *SessionIdData, backend *Backend, hello *HelloClientMessage, auth *BackendClientAuthResponse) (*ClientSession, error) {
 	s := &ClientSession{
 		hub:       hub,
+		events:    hub.events,
 		privateId: privateId,
 		publicId:  publicId,
 		data:      data,
@@ -106,14 +105,13 @@ func NewClientSession(hub *Hub, privateId string, publicId string, data *Session
 		userData:   auth.User,
 
 		backend: backend,
-
-		natsReceiver: make(chan *nats.Msg, 64),
-		stopRun:      make(chan bool, 1),
-		runStopped:   make(chan bool, 1),
 	}
 	if s.clientType == HelloClientTypeInternal {
 		s.backendUrl = hello.Auth.internalParams.Backend
 		s.parsedBackendUrl = hello.Auth.internalParams.parsedBackend
+		if !s.HasFeature(ClientFeatureInternalInCall) {
+			s.SetInCall(FlagInCall | FlagWithAudio)
+		}
 	} else {
 		s.backendUrl = hello.Auth.Url
 		s.parsedBackendUrl = hello.Auth.parsedUrl
@@ -124,23 +122,22 @@ func NewClientSession(hub *Hub, privateId string, publicId string, data *Session
 			backendUrl += "/"
 		}
 		backendUrl += PathToOcsSignalingBackend
-		if u, err := url.Parse(backendUrl); err != nil {
+		u, err := url.Parse(backendUrl)
+		if err != nil {
 			return nil, err
-		} else {
-			if strings.Contains(u.Host, ":") && hasStandardPort(u) {
-				u.Host = u.Hostname()
-			}
-
-			s.backendUrl = backendUrl
-			s.parsedBackendUrl = u
 		}
+
+		if strings.Contains(u.Host, ":") && hasStandardPort(u) {
+			u.Host = u.Hostname()
+		}
+
+		s.backendUrl = backendUrl
+		s.parsedBackendUrl = u
 	}
 
-	if err := s.SubscribeNats(hub.nats); err != nil {
+	if err := s.SubscribeEvents(); err != nil {
 		return nil, err
 	}
-	atomic.StoreInt32(&s.running, 1)
-	go s.run()
 	return s, nil
 }
 
@@ -153,6 +150,8 @@ func (s *ClientSession) PublicId() string {
 }
 
 func (s *ClientSession) RoomSessionId() string {
+	s.mu.Lock()
+	defer s.mu.Unlock()
 	return s.roomSessionId
 }
 
@@ -164,6 +163,28 @@ func (s *ClientSession) ClientType() string {
 	return s.clientType
 }
 
+// GetInCall is only used for internal clients.
+func (s *ClientSession) GetInCall() int {
+	return int(atomic.LoadUint32(&s.inCall))
+}
+
+func (s *ClientSession) SetInCall(inCall int) bool {
+	if inCall < 0 {
+		inCall = 0
+	}
+
+	for {
+		old := atomic.LoadUint32(&s.inCall)
+		if old == uint32(inCall) {
+			return false
+		}
+
+		if atomic.CompareAndSwapUint32(&s.inCall, old, uint32(inCall)) {
+			return true
+		}
+	}
+}
+
 func (s *ClientSession) GetFeatures() []string {
 	return s.features
 }
@@ -213,6 +234,9 @@ func (s *ClientSession) hasAnyPermissionLocked(permission ...Permission) bool {
 func (s *ClientSession) hasPermissionLocked(permission Permission) bool {
 	if !s.supportsPermissions {
 		// Old-style session that doesn't receive permissions from Nextcloud.
+		if result, found := DefaultPermissionOverrides[permission]; found {
+			return result
+		}
 		return true
 	}
 
@@ -294,19 +318,6 @@ func (s *ClientSession) UserData() *json.RawMessage {
 	return s.userData
 }
 
-func (s *ClientSession) run() {
-loop:
-	for {
-		select {
-		case msg := <-s.natsReceiver:
-			s.processClientMessage(msg)
-		case <-s.stopRun:
-			break loop
-		}
-	}
-	s.runStopped <- true
-}
-
 func (s *ClientSession) StartExpire() {
 	// The hub mutex must be held when calling this method.
 	s.expires = time.Now().Add(sessionExpireDuration)
@@ -329,6 +340,10 @@ func (s *ClientSession) SetRoom(room *Room) {
 	} else {
 		atomic.StoreInt64(&s.roomJoinTime, 0)
 	}
+
+	s.seenJoinedLock.Lock()
+	defer s.seenJoinedLock.Unlock()
+	s.seenJoinedEvents = nil
 }
 
 func (s *ClientSession) GetRoom() *Room {
@@ -336,7 +351,12 @@ func (s *ClientSession) GetRoom() *Room {
 }
 
 func (s *ClientSession) getRoomJoinTime() time.Time {
-	return time.Unix(0, atomic.LoadInt64(&s.roomJoinTime))
+	t := atomic.LoadInt64(&s.roomJoinTime)
+	if t == 0 {
+		return time.Time{}
+	}
+
+	return time.Unix(0, t)
 }
 
 func (s *ClientSession) releaseMcuObjects() {
@@ -369,18 +389,10 @@ func (s *ClientSession) closeAndWait(wait bool) {
 
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	if s.userSubscription != nil {
-		if err := s.userSubscription.Unsubscribe(); err != nil {
-			log.Printf("Error closing user subscription in session %s: %s", s.PublicId(), err)
-		}
-		s.userSubscription = nil
-	}
-	if s.sessionSubscription != nil {
-		if err := s.sessionSubscription.Unsubscribe(); err != nil {
-			log.Printf("Error closing session subscription in session %s: %s", s.PublicId(), err)
-		}
-		s.sessionSubscription = nil
+	if s.userId != "" {
+		s.events.UnregisterUserListener(s.userId, s.backend, s)
 	}
+	s.events.UnregisterSessionListener(s.publicId, s.backend, s)
 	go func(virtualSessions map[*VirtualSession]bool) {
 		for session := range virtualSessions {
 			session.Close()
@@ -390,56 +402,32 @@ func (s *ClientSession) closeAndWait(wait bool) {
 	s.releaseMcuObjects()
 	s.clearClientLocked(nil)
 	s.backend.RemoveSession(s)
-	if atomic.CompareAndSwapInt32(&s.running, 1, 0) {
-		s.stopRun <- true
-		// Only wait if called from outside the Session goroutine.
-		if wait {
-			s.mu.Unlock()
-			// Wait for Session goroutine to stop
-			<-s.runStopped
-			s.mu.Lock()
-		}
-	}
 }
 
-func GetSubjectForUserId(userId string, backend *Backend) string {
-	if backend == nil || backend.IsCompat() {
-		return GetEncodedSubject("user", userId)
-	} else {
-		return GetEncodedSubject("user", userId+"|"+backend.Id())
-	}
-}
-
-func (s *ClientSession) SubscribeNats(n NatsClient) error {
+func (s *ClientSession) SubscribeEvents() error {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	var err error
 	if s.userId != "" {
-		if s.userSubscription, err = n.Subscribe(GetSubjectForUserId(s.userId, s.backend), s.natsReceiver); err != nil {
+		if err := s.events.RegisterUserListener(s.userId, s.backend, s); err != nil {
 			return err
 		}
 	}
 
-	if s.sessionSubscription, err = n.Subscribe("session."+s.publicId, s.natsReceiver); err != nil {
-		return err
-	}
-
-	return nil
+	return s.events.RegisterSessionListener(s.publicId, s.backend, s)
 }
 
-func (s *ClientSession) SubscribeRoomNats(n NatsClient, roomid string, roomSessionId string) error {
+func (s *ClientSession) SubscribeRoomEvents(roomid string, roomSessionId string) error {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	var err error
-	if s.roomSubscription, err = n.Subscribe(GetSubjectForRoomId(roomid, s.Backend()), s.natsReceiver); err != nil {
+	if err := s.events.RegisterRoomListener(roomid, s.backend, s); err != nil {
 		return err
 	}
 
 	if roomSessionId != "" {
-		if err = s.hub.roomSessions.SetRoomSession(s, roomSessionId); err != nil {
-			s.doUnsubscribeRoomNats(true)
+		if err := s.hub.roomSessions.SetRoomSession(s, roomSessionId); err != nil {
+			s.doUnsubscribeRoomEvents(true)
 			return err
 		}
 	}
@@ -470,29 +458,26 @@ func (s *ClientSession) LeaveRoom(notify bool) *Room {
 		return nil
 	}
 
-	s.doUnsubscribeRoomNats(notify)
+	s.doUnsubscribeRoomEvents(notify)
 	s.SetRoom(nil)
 	s.releaseMcuObjects()
 	room.RemoveSession(s)
 	return room
 }
 
-func (s *ClientSession) UnsubscribeRoomNats() {
+func (s *ClientSession) UnsubscribeRoomEvents() {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	s.doUnsubscribeRoomNats(true)
+	s.doUnsubscribeRoomEvents(true)
 }
 
-func (s *ClientSession) doUnsubscribeRoomNats(notify bool) {
-	if s.roomSubscription != nil {
-		if err := s.roomSubscription.Unsubscribe(); err != nil {
-			log.Printf("Error closing room subscription in session %s: %s", s.PublicId(), err)
-		}
-		s.roomSubscription = nil
+func (s *ClientSession) doUnsubscribeRoomEvents(notify bool) {
+	room := s.GetRoom()
+	if room != nil {
+		s.events.UnregisterRoomListener(room.Id(), s.Backend(), s)
 	}
 	s.hub.roomSessions.DeleteRoomSession(s)
-	room := s.GetRoom()
 	if notify && room != nil && s.roomSessionId != "" {
 		// Notify
 		go func(sid string) {
@@ -563,6 +548,34 @@ func (s *ClientSession) SetClient(client *Client) *Client {
 	return prev
 }
 
+func (s *ClientSession) sendOffer(client McuClient, sender string, streamType string, offer map[string]interface{}) {
+	offer_message := &AnswerOfferMessage{
+		To:       s.PublicId(),
+		From:     sender,
+		Type:     "offer",
+		RoomType: streamType,
+		Payload:  offer,
+		Sid:      client.Sid(),
+	}
+	offer_data, err := json.Marshal(offer_message)
+	if err != nil {
+		log.Println("Could not serialize offer", offer_message, err)
+		return
+	}
+	response_message := &ServerMessage{
+		Type: "message",
+		Message: &MessageServerMessage{
+			Sender: &MessageServerMessageSender{
+				Type:      "session",
+				SessionId: sender,
+			},
+			Data: (*json.RawMessage)(&offer_data),
+		},
+	}
+
+	s.sendMessageUnlocked(response_message)
+}
+
 func (s *ClientSession) sendCandidate(client McuClient, sender string, streamType string, candidate interface{}) {
 	candidate_message := &AnswerOfferMessage{
 		To:       s.PublicId(),
@@ -572,6 +585,7 @@ func (s *ClientSession) sendCandidate(client McuClient, sender string, streamTyp
 		Payload: map[string]interface{}{
 			"candidate": candidate,
 		},
+		Sid: client.Sid(),
 	}
 	candidate_data, err := json.Marshal(candidate_message)
 	if err != nil {
@@ -612,6 +626,11 @@ func (s *ClientSession) SendError(e *Error) bool {
 }
 
 func (s *ClientSession) SendMessage(message *ServerMessage) bool {
+	message = s.filterMessage(message)
+	if message == nil {
+		return true
+	}
+
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
@@ -628,6 +647,18 @@ func (s *ClientSession) SendMessages(messages []*ServerMessage) bool {
 	return true
 }
 
+func (s *ClientSession) OnUpdateOffer(client McuClient, offer map[string]interface{}) {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	for _, sub := range s.subscribers {
+		if sub.Id() == client.Id() {
+			s.sendOffer(client, sub.Publisher(), client.StreamType(), offer)
+			return
+		}
+	}
+}
+
 func (s *ClientSession) OnIceCandidate(client McuClient, candidate interface{}) {
 	s.mu.Lock()
 	defer s.mu.Unlock()
@@ -657,6 +688,9 @@ func (s *ClientSession) OnIceCompleted(client McuClient) {
 	// s.OnIceCandidate(client, nil)
 }
 
+func (s *ClientSession) SubscriberSidUpdated(subscriber McuSubscriber) {
+}
+
 func (s *ClientSession) PublisherClosed(publisher McuPublisher) {
 	s.mu.Lock()
 	defer s.mu.Unlock()
@@ -720,7 +754,7 @@ func (s *ClientSession) isSdpAllowedToSendLocked(payload map[string]interface{})
 		return 0, &SdpError{"payload does not contain a valid sdp"}
 	}
 	var sdp sdp.SessionDescription
-	if err := sdp.Unmarshal(sdpText); err != nil {
+	if err := sdp.Unmarshal([]byte(sdpText)); err != nil {
 		return 0, &WrappedSdpError{
 			SdpError: SdpError{
 				message: fmt.Sprintf("could not parse sdp: %s", err),
@@ -819,6 +853,7 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 	if !found {
 		client := s.getClientUnlocked()
 		s.mu.Unlock()
+		defer s.mu.Lock()
 
 		bitrate := data.Bitrate
 		if backend := s.Backend(); backend != nil {
@@ -835,8 +870,7 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 			}
 		}
 		var err error
-		publisher, err = mcu.NewPublisher(ctx, s, s.PublicId(), streamType, bitrate, mediaTypes, client)
-		s.mu.Lock()
+		publisher, err = mcu.NewPublisher(ctx, s, s.PublicId(), data.Sid, streamType, bitrate, mediaTypes, client)
 		if err != nil {
 			return nil, err
 		}
@@ -854,16 +888,52 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 			s.publishers[streamType] = publisher
 		}
 		log.Printf("Publishing %s as %s for session %s", streamType, publisher.Id(), s.PublicId())
+		s.publisherWaiters.Wakeup()
+	} else {
+		publisher.SetMedia(mediaTypes)
 	}
 
 	return publisher, nil
 }
 
+func (s *ClientSession) getPublisherLocked(streamType string) McuPublisher {
+	return s.publishers[streamType]
+}
+
 func (s *ClientSession) GetPublisher(streamType string) McuPublisher {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	return s.publishers[streamType]
+	return s.getPublisherLocked(streamType)
+}
+
+func (s *ClientSession) GetOrWaitForPublisher(ctx context.Context, streamType string) McuPublisher {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	publisher := s.getPublisherLocked(streamType)
+	if publisher != nil {
+		return publisher
+	}
+
+	ch := make(chan struct{}, 1)
+	id := s.publisherWaiters.Add(ch)
+	defer s.publisherWaiters.Remove(id)
+
+	for {
+		s.mu.Unlock()
+		select {
+		case <-ch:
+			s.mu.Lock()
+			publisher := s.getPublisherLocked(streamType)
+			if publisher != nil {
+				return publisher
+			}
+		case <-ctx.Done():
+			s.mu.Lock()
+			return nil
+		}
+	}
 }
 
 func (s *ClientSession) GetOrCreateSubscriber(ctx context.Context, mcu Mcu, id string, streamType string) (McuSubscriber, error) {
@@ -907,13 +977,19 @@ func (s *ClientSession) GetSubscriber(id string, streamType string) McuSubscribe
 	return s.subscribers[id+"|"+streamType]
 }
 
-func (s *ClientSession) processClientMessage(msg *nats.Msg) {
-	var message NatsMessage
-	if err := s.hub.nats.Decode(msg, &message); err != nil {
-		log.Printf("Could not decode NATS message %+v for session %s: %s", *msg, s.PublicId(), err)
-		return
-	}
+func (s *ClientSession) ProcessAsyncRoomMessage(message *AsyncMessage) {
+	s.processAsyncMessage(message)
+}
 
+func (s *ClientSession) ProcessAsyncUserMessage(message *AsyncMessage) {
+	s.processAsyncMessage(message)
+}
+
+func (s *ClientSession) ProcessAsyncSessionMessage(message *AsyncMessage) {
+	s.processAsyncMessage(message)
+}
+
+func (s *ClientSession) processAsyncMessage(message *AsyncMessage) {
 	switch message.Type {
 	case "permissions":
 		s.SetPermissions(message.Permissions)
@@ -948,16 +1024,75 @@ func (s *ClientSession) processClientMessage(msg *nats.Msg) {
 		return
 	case "message":
 		if message.Message.Type == "bye" && message.Message.Bye.Reason == "room_session_reconnected" {
-			s.mu.Lock()
-			roomSessionId := s.RoomSessionId()
-			s.mu.Unlock()
-			log.Printf("Closing session %s because same room session %s connected", s.PublicId(), roomSessionId)
+			log.Printf("Closing session %s because same room session %s connected", s.PublicId(), s.RoomSessionId())
 			s.LeaveRoom(false)
 			defer s.closeAndWait(false)
 		}
+	case "sendoffer":
+		// Process asynchronously to not block other messages received.
+		go func() {
+			ctx, cancel := context.WithTimeout(context.Background(), s.hub.mcuTimeout)
+			defer cancel()
+
+			mc, err := s.GetOrCreateSubscriber(ctx, s.hub.mcu, message.SendOffer.SessionId, message.SendOffer.Data.RoomType)
+			if err != nil {
+				log.Printf("Could not create MCU subscriber for session %s to process sendoffer in %s: %s", message.SendOffer.SessionId, s.PublicId(), err)
+				if err := s.events.PublishSessionMessage(message.SendOffer.SessionId, s.backend, &AsyncMessage{
+					Type: "message",
+					Message: &ServerMessage{
+						Id:    message.SendOffer.MessageId,
+						Type:  "error",
+						Error: NewError("client_not_found", "No MCU client found to send message to."),
+					},
+				}); err != nil {
+					log.Printf("Error sending sendoffer error response to %s: %s", message.SendOffer.SessionId, err)
+				}
+				return
+			} else if mc == nil {
+				log.Printf("No MCU subscriber found for session %s to process sendoffer in %s", message.SendOffer.SessionId, s.PublicId())
+				if err := s.events.PublishSessionMessage(message.SendOffer.SessionId, s.backend, &AsyncMessage{
+					Type: "message",
+					Message: &ServerMessage{
+						Id:    message.SendOffer.MessageId,
+						Type:  "error",
+						Error: NewError("client_not_found", "No MCU client found to send message to."),
+					},
+				}); err != nil {
+					log.Printf("Error sending sendoffer error response to %s: %s", message.SendOffer.SessionId, err)
+				}
+				return
+			}
+
+			mc.SendMessage(context.TODO(), nil, message.SendOffer.Data, func(err error, response map[string]interface{}) {
+				if err != nil {
+					log.Printf("Could not send MCU message %+v for session %s to %s: %s", message.SendOffer.Data, message.SendOffer.SessionId, s.PublicId(), err)
+					if err := s.events.PublishSessionMessage(message.SendOffer.SessionId, s.backend, &AsyncMessage{
+						Type: "message",
+						Message: &ServerMessage{
+							Id:    message.SendOffer.MessageId,
+							Type:  "error",
+							Error: NewError("processing_failed", "Processing of the message failed, please check server logs."),
+						},
+					}); err != nil {
+						log.Printf("Error sending sendoffer error response to %s: %s", message.SendOffer.SessionId, err)
+					}
+					return
+				} else if response == nil {
+					// No response received
+					return
+				}
+
+				s.hub.sendMcuMessageResponse(s, mc, &MessageClientMessage{
+					Recipient: MessageClientMessageRecipient{
+						SessionId: message.SendOffer.SessionId,
+					},
+				}, message.SendOffer.Data, response)
+			})
+		}()
+		return
 	}
 
-	serverMessage := s.processNatsMessage(&message)
+	serverMessage := s.filterAsyncMessage(message)
 	if serverMessage == nil {
 		return
 	}
@@ -983,11 +1118,188 @@ func (s *ClientSession) storePendingMessage(message *ServerMessage) {
 	}
 }
 
-func (s *ClientSession) processNatsMessage(msg *NatsMessage) *ServerMessage {
+func filterDisplayNames(events []*EventServerMessageSessionEntry) []*EventServerMessageSessionEntry {
+	result := make([]*EventServerMessageSessionEntry, 0, len(events))
+	for _, event := range events {
+		if event.User == nil {
+			result = append(result, event)
+			continue
+		}
+
+		var userdata map[string]interface{}
+		if err := json.Unmarshal(*event.User, &userdata); err != nil {
+			result = append(result, event)
+			continue
+		}
+
+		if _, found := userdata["displayname"]; !found {
+			result = append(result, event)
+			continue
+		}
+
+		delete(userdata, "displayname")
+		if len(userdata) == 0 {
+			// No more userdata, no need to serialize empty map.
+			e := event.Clone()
+			e.User = nil
+			result = append(result, e)
+			continue
+		}
+
+		data, err := json.Marshal(userdata)
+		if err != nil {
+			result = append(result, event)
+			continue
+		}
+
+		e := event.Clone()
+		e.User = (*json.RawMessage)(&data)
+		result = append(result, e)
+	}
+	return result
+}
+
+func (s *ClientSession) filterDuplicateJoin(entries []*EventServerMessageSessionEntry) []*EventServerMessageSessionEntry {
+	s.seenJoinedLock.Lock()
+	defer s.seenJoinedLock.Unlock()
+
+	// Due to the asynchronous events, a session might received a "Joined" event
+	// for the same (other) session twice, so filter these out on a per-session
+	// level.
+	result := make([]*EventServerMessageSessionEntry, 0, len(entries))
+	for _, e := range entries {
+		if s.seenJoinedEvents[e.SessionId] {
+			log.Printf("Session %s got duplicate joined event for %s, ignoring", s.publicId, e.SessionId)
+			continue
+		}
+
+		if s.seenJoinedEvents == nil {
+			s.seenJoinedEvents = make(map[string]bool)
+		}
+		s.seenJoinedEvents[e.SessionId] = true
+		result = append(result, e)
+	}
+	return result
+}
+
+func (s *ClientSession) filterMessage(message *ServerMessage) *ServerMessage {
+	switch message.Type {
+	case "event":
+		switch message.Event.Target {
+		case "participants":
+			if message.Event.Type == "update" {
+				m := message.Event.Update
+				users := make(map[string]bool)
+				for _, entry := range m.Users {
+					users[entry["sessionId"].(string)] = true
+				}
+				for _, entry := range m.Changed {
+					if users[entry["sessionId"].(string)] {
+						continue
+					}
+					m.Users = append(m.Users, entry)
+				}
+				// TODO(jojo): Only send all users if current session id has
+				// changed its "inCall" flag to true.
+				m.Changed = nil
+			}
+		case "room":
+			switch message.Event.Type {
+			case "join":
+				join := s.filterDuplicateJoin(message.Event.Join)
+				if len(join) == 0 {
+					return nil
+				}
+				copied := false
+				if len(join) != len(message.Event.Join) {
+					// Create unique copy of message for only this client.
+					copied = true
+					message = &ServerMessage{
+						Id:   message.Id,
+						Type: message.Type,
+						Event: &EventServerMessage{
+							Type:   message.Event.Type,
+							Target: message.Event.Target,
+							Join:   join,
+						},
+					}
+				}
+
+				if s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
+					if copied {
+						message.Event.Join = filterDisplayNames(message.Event.Join)
+					} else {
+						message = &ServerMessage{
+							Id:   message.Id,
+							Type: message.Type,
+							Event: &EventServerMessage{
+								Type:   message.Event.Type,
+								Target: message.Event.Target,
+								Join:   filterDisplayNames(message.Event.Join),
+							},
+						}
+					}
+				}
+			case "leave":
+				s.seenJoinedLock.Lock()
+				defer s.seenJoinedLock.Unlock()
+
+				for _, e := range message.Event.Leave {
+					delete(s.seenJoinedEvents, e)
+				}
+			case "message":
+				if message.Event.Message == nil || message.Event.Message.Data == nil || len(*message.Event.Message.Data) == 0 || !s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
+					return message
+				}
+
+				var data RoomEventMessageData
+				if err := json.Unmarshal(*message.Event.Message.Data, &data); err != nil {
+					return message
+				}
+
+				if data.Type == "chat" && data.Chat != nil && data.Chat.Comment != nil {
+					if displayName, found := (*data.Chat.Comment)["actorDisplayName"]; found && displayName != "" {
+						(*data.Chat.Comment)["actorDisplayName"] = ""
+						if encoded, err := json.Marshal(data); err == nil {
+							// Create unique copy of message for only this client.
+							message = &ServerMessage{
+								Id:   message.Id,
+								Type: message.Type,
+								Event: &EventServerMessage{
+									Type:   message.Event.Type,
+									Target: message.Event.Target,
+									Message: &RoomEventMessage{
+										RoomId: message.Event.Message.RoomId,
+										Data:   (*json.RawMessage)(&encoded),
+									},
+								},
+							}
+						}
+					}
+				}
+			}
+		}
+	case "message":
+		if message.Message != nil && message.Message.Data != nil && len(*message.Message.Data) > 0 && s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
+			var data MessageServerMessageData
+			if err := json.Unmarshal(*message.Message.Data, &data); err != nil {
+				return message
+			}
+
+			if data.Type == "nickChanged" {
+				return nil
+			}
+		}
+	}
+
+	return message
+}
+
+func (s *ClientSession) filterAsyncMessage(msg *AsyncMessage) *ServerMessage {
 	switch msg.Type {
 	case "message":
 		if msg.Message == nil {
-			log.Printf("Received NATS message without payload: %+v", msg)
+			log.Printf("Received asynchronous message without payload: %+v", msg)
 			return nil
 		}
 
@@ -1007,27 +1319,11 @@ func (s *ClientSession) processNatsMessage(msg *NatsMessage) *ServerMessage {
 				return nil
 			}
 		case "event":
-			if msg.Message.Event.Target == "participants" &&
-				msg.Message.Event.Type == "update" {
-				m := msg.Message.Event.Update
-				users := make(map[string]bool)
-				for _, entry := range m.Users {
-					users[entry["sessionId"].(string)] = true
-				}
-				for _, entry := range m.Changed {
-					if users[entry["sessionId"].(string)] {
-						continue
-					}
-					m.Users = append(m.Users, entry)
-				}
-				// TODO(jojo): Only send all users if current session id has
-				// changed its "inCall" flag to true.
-				m.Changed = nil
-			} else if msg.Message.Event.Target == "room" {
-				// Can happen mostly during tests where an older room NATS message
+			if msg.Message.Event.Target == "room" {
+				// Can happen mostly during tests where an older room async message
 				// could be received by a subscriber that joined after it was sent.
-				if msg.SendTime.Before(s.getRoomJoinTime()) {
-					log.Printf("Message %+v was sent before room was joined, ignoring", msg.Message)
+				if joined := s.getRoomJoinTime(); joined.IsZero() || msg.SendTime.Before(joined) {
+					log.Printf("Message %+v was sent on %s before room was joined on %s, ignoring", msg.Message, msg.SendTime, joined)
 					return nil
 				}
 			}
@@ -1035,7 +1331,7 @@ func (s *ClientSession) processNatsMessage(msg *NatsMessage) *ServerMessage {
 
 		return msg.Message
 	default:
-		log.Printf("Received NATS message with unsupported type %s: %+v", msg.Type, msg)
+		log.Printf("Received async message with unsupported type %s: %+v", msg.Type, msg)
 		return nil
 	}
 }
@@ -1083,3 +1379,14 @@ func (s *ClientSession) RemoveVirtualSession(session *VirtualSession) {
 	delete(s.virtualSessions, session)
 	s.mu.Unlock()
 }
+
+func (s *ClientSession) GetVirtualSessions() []*VirtualSession {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	result := make([]*VirtualSession, 0, len(s.virtualSessions))
+	for session := range s.virtualSessions {
+		result = append(result, session)
+	}
+	return result
+}
