@@ -39,7 +39,7 @@ import (
 	"time"
 
 	"github.com/dlintw/goconf"
-	"github.com/golang-jwt/jwt"
+	"github.com/golang-jwt/jwt/v4"
 	"github.com/google/uuid"
 	"github.com/gorilla/mux"
 	"github.com/gorilla/securecookie"
@@ -73,6 +73,7 @@ var (
 	TimeoutCreatingSubscriber = signaling.NewError("timeout", "Timeout creating subscriber.")
 	TokenAuthFailed           = signaling.NewError("auth_failed", "The token could not be authenticated.")
 	TokenExpired              = signaling.NewError("token_expired", "The token is expired.")
+	TokenNotValidYet          = signaling.NewError("token_not_valid_yet", "The token is not valid yet.")
 	UnknownClient             = signaling.NewError("unknown_client", "Unknown client id given.")
 	UnsupportedCommand        = signaling.NewError("bad_request", "Unsupported command received.")
 	UnsupportedMessage        = signaling.NewError("bad_request", "Unsupported message received.")
@@ -91,13 +92,13 @@ type ProxyServer struct {
 	mcu     signaling.Mcu
 	stopped uint32
 
-	shutdownChannel   chan bool
+	shutdownChannel   chan struct{}
 	shutdownScheduled uint32
 
 	upgrader websocket.Upgrader
 
 	tokens          ProxyTokens
-	statsAllowedIps map[string]bool
+	statsAllowedIps *signaling.AllowedIps
 
 	sid          uint64
 	cookie       *securecookie.SecureCookie
@@ -140,21 +141,16 @@ func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*
 	}
 
 	statsAllowed, _ := config.GetString("stats", "allowed_ips")
-	var statsAllowedIps map[string]bool
-	if statsAllowed == "" {
-		log.Printf("No IPs configured for the stats endpoint, only allowing access from 127.0.0.1")
-		statsAllowedIps = map[string]bool{
-			"127.0.0.1": true,
-		}
+	statsAllowedIps, err := signaling.ParseAllowedIps(statsAllowed)
+	if err != nil {
+		return nil, err
+	}
+
+	if !statsAllowedIps.Empty() {
+		log.Printf("Only allowing access to the stats endpoint from %s", statsAllowed)
 	} else {
-		log.Printf("Only allowing access to the stats endpoing from %s", statsAllowed)
-		statsAllowedIps = make(map[string]bool)
-		for _, ip := range strings.Split(statsAllowed, ",") {
-			ip = strings.TrimSpace(ip)
-			if ip != "" {
-				statsAllowedIps[ip] = true
-			}
-		}
+		log.Printf("No IPs configured for the stats endpoint, only allowing access from 127.0.0.1")
+		statsAllowedIps = signaling.DefaultAllowedIps()
 	}
 
 	country, _ := config.GetString("app", "country")
@@ -171,7 +167,7 @@ func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*
 		version: version,
 		country: country,
 
-		shutdownChannel: make(chan bool, 1),
+		shutdownChannel: make(chan struct{}),
 
 		upgrader: websocket.Upgrader{
 			ReadBufferSize:  websocketReadBufferSize,
@@ -357,11 +353,13 @@ func (s *ProxyServer) Stop() {
 		return
 	}
 
-	s.mcu.Stop()
+	if s.mcu != nil {
+		s.mcu.Stop()
+	}
 	s.tokens.Close()
 }
 
-func (s *ProxyServer) ShutdownChannel() chan bool {
+func (s *ProxyServer) ShutdownChannel() <-chan struct{} {
 	return s.shutdownChannel
 }
 
@@ -381,9 +379,7 @@ func (s *ProxyServer) ScheduleShutdown() {
 	})
 
 	if s.GetClientCount() == 0 {
-		go func() {
-			s.shutdownChannel <- true
-		}()
+		go close(s.shutdownChannel)
 	}
 }
 
@@ -430,21 +426,6 @@ func (s *ProxyServer) proxyHandler(w http.ResponseWriter, r *http.Request) {
 		return
 	}
 
-	client.OnClosed = func(c *signaling.Client) {
-		if session := client.GetSession(); session != nil {
-			session.MarkUsed()
-		}
-		s.clientClosed(c)
-	}
-	client.OnMessageReceived = func(c *signaling.Client, data []byte) {
-		s.processMessage(client, data)
-	}
-	client.OnRTTReceived = func(c *signaling.Client, rtt time.Duration) {
-		if session := client.GetSession(); session != nil {
-			session.MarkUsed()
-		}
-	}
-
 	go client.WritePump()
 	go client.ReadPump()
 }
@@ -586,9 +567,9 @@ func (s *ProxyServer) processMessage(client *ProxyClient, data []byte) {
 			Id:   message.Id,
 			Type: "hello",
 			Hello: &signaling.HelloProxyServerMessage{
-				Version:   signaling.HelloVersion,
+				Version:   signaling.HelloVersionV1,
 				SessionId: session.PublicId(),
-				Server: &signaling.HelloServerMessageServer{
+				Server: &signaling.WelcomeServerMessage{
 					Version: s.version,
 					Country: s.country,
 				},
@@ -635,7 +616,7 @@ func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, s
 		}
 
 		id := uuid.New().String()
-		publisher, err := s.mcu.NewPublisher(ctx, session, id, cmd.StreamType, cmd.Bitrate, cmd.MediaTypes, &emptyInitiator{})
+		publisher, err := s.mcu.NewPublisher(ctx, session, id, cmd.Sid, cmd.StreamType, cmd.Bitrate, cmd.MediaTypes, &emptyInitiator{})
 		if err == context.DeadlineExceeded {
 			log.Printf("Timeout while creating %s publisher %s for %s", cmd.StreamType, id, session.PublicId())
 			session.sendMessage(message.NewErrorServerMessage(TimeoutCreatingPublisher))
@@ -682,7 +663,8 @@ func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, s
 			Id:   message.Id,
 			Type: "command",
 			Command: &signaling.CommandProxyServerMessage{
-				Id: id,
+				Id:  id,
+				Sid: subscriber.Sid(),
 			},
 		}
 		session.sendMessage(response)
@@ -785,6 +767,7 @@ func (s *ProxyServer) processPayload(ctx context.Context, client *ProxyClient, s
 	case "candidate":
 		mcuData = &signaling.MessageClientMessageData{
 			Type:    payload.Type,
+			Sid:     payload.Sid,
 			Payload: payload.Payload,
 		}
 	case "endOfCandidates":
@@ -803,6 +786,7 @@ func (s *ProxyServer) processPayload(ctx context.Context, client *ProxyClient, s
 	case "sendoffer":
 		mcuData = &signaling.MessageClientMessageData{
 			Type: payload.Type,
+			Sid:  payload.Sid,
 		}
 	default:
 		session.sendMessage(message.NewErrorServerMessage(UnsupportedPayload))
@@ -863,8 +847,15 @@ func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*Pro
 			reason = "unsupported-issuer"
 			return nil, fmt.Errorf("No key found for issuer")
 		}
+
 		return tokenKey.key, nil
 	})
+	if err, ok := err.(*jwt.ValidationError); ok {
+		if err.Errors&jwt.ValidationErrorIssuedAt == jwt.ValidationErrorIssuedAt {
+			statsTokenErrorsTotal.WithLabelValues("not-valid-yet").Inc()
+			return nil, TokenNotValidYet
+		}
+	}
 	if err != nil {
 		statsTokenErrorsTotal.WithLabelValues(reason).Inc()
 		return nil, TokenAuthFailed
@@ -877,7 +868,7 @@ func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*Pro
 	}
 
 	minIssuedAt := time.Now().Add(-maxTokenAge)
-	if issuedAt := time.Unix(claims.IssuedAt, 0); issuedAt.Before(minIssuedAt) {
+	if issuedAt := claims.IssuedAt; issuedAt != nil && issuedAt.Before(minIssuedAt) {
 		statsTokenErrorsTotal.WithLabelValues("expired").Inc()
 		return nil, TokenExpired
 	}
@@ -964,9 +955,7 @@ func (s *ProxyServer) DeleteClient(id string, client signaling.McuClient) bool {
 	delete(s.clientIds, client.Id())
 
 	if len(s.clients) == 0 && atomic.LoadUint32(&s.shutdownScheduled) != 0 {
-		go func() {
-			s.shutdownChannel <- true
-		}()
+		go close(s.shutdownChannel)
 	}
 	return true
 }
@@ -998,15 +987,25 @@ func (s *ProxyServer) getStats() map[string]interface{} {
 	return result
 }
 
+func (s *ProxyServer) allowStatsAccess(r *http.Request) bool {
+	addr := getRealUserIP(r)
+	if strings.Contains(addr, ":") {
+		if host, _, err := net.SplitHostPort(addr); err == nil {
+			addr = host
+		}
+	}
+
+	ip := net.ParseIP(addr)
+	if ip == nil {
+		return false
+	}
+
+	return s.statsAllowedIps.Allowed(ip)
+}
+
 func (s *ProxyServer) validateStatsRequest(f func(http.ResponseWriter, *http.Request)) func(http.ResponseWriter, *http.Request) {
 	return func(w http.ResponseWriter, r *http.Request) {
-		addr := getRealUserIP(r)
-		if strings.Contains(addr, ":") {
-			if host, _, err := net.SplitHostPort(addr); err == nil {
-				addr = host
-			}
-		}
-		if !s.statsAllowedIps[addr] {
+		if !s.allowStatsAccess(r) {
 			http.Error(w, "Authentication check failed", http.StatusForbidden)
 			return
 		}
