@@ -28,10 +28,10 @@ import (
 	"fmt"
 	"log"
 	"net/url"
+	"strconv"
 	"sync"
 	"time"
 
-	"github.com/nats-io/nats.go"
 	"github.com/prometheus/client_golang/prometheus"
 )
 
@@ -44,6 +44,13 @@ const (
 	FlagWithPhone    = 8
 )
 
+type SessionChangeFlag int
+
+const (
+	SessionChangeFlags  SessionChangeFlag = 1
+	SessionChangeInCall SessionChangeFlag = 2
+)
+
 var (
 	updateActiveSessionsInterval = 10 * time.Second
 )
@@ -55,14 +62,14 @@ func init() {
 type Room struct {
 	id      string
 	hub     *Hub
-	nats    NatsClient
+	events  AsyncEvents
 	backend *Backend
 
 	properties *json.RawMessage
 
-	closeChan chan bool
-	mu        *sync.RWMutex
-	sessions  map[string]Session
+	closer   *Closer
+	mu       *sync.RWMutex
+	sessions map[string]Session
 
 	internalSessions map[Session]bool
 	virtualSessions  map[*VirtualSession]bool
@@ -71,30 +78,13 @@ type Room struct {
 
 	statsRoomSessionsCurrent *prometheus.GaugeVec
 
-	natsReceiver        chan *nats.Msg
-	backendSubscription NatsSubscription
-
 	// Users currently in the room
 	users []map[string]interface{}
 
-	// Timestamps of last NATS backend requests for the different types.
-	lastNatsRoomRequests map[string]int64
-}
+	// Timestamps of last backend requests for the different types.
+	lastRoomRequests map[string]int64
 
-func GetSubjectForRoomId(roomId string, backend *Backend) string {
-	if backend == nil || backend.IsCompat() {
-		return GetEncodedSubject("room", roomId)
-	} else {
-		return GetEncodedSubject("room", roomId+"|"+backend.Id())
-	}
-}
-
-func GetSubjectForBackendRoomId(roomId string, backend *Backend) string {
-	if backend == nil || backend.IsCompat() {
-		return GetEncodedSubject("backend.room", roomId)
-	} else {
-		return GetEncodedSubject("backend.room", roomId+"|"+backend.Id())
-	}
+	transientData *TransientData
 }
 
 func getRoomIdForBackend(id string, backend *Backend) string {
@@ -105,25 +95,18 @@ func getRoomIdForBackend(id string, backend *Backend) string {
 	return backend.Id() + "|" + id
 }
 
-func NewRoom(roomId string, properties *json.RawMessage, hub *Hub, n NatsClient, backend *Backend) (*Room, error) {
-	natsReceiver := make(chan *nats.Msg, 64)
-	backendSubscription, err := n.Subscribe(GetSubjectForBackendRoomId(roomId, backend), natsReceiver)
-	if err != nil {
-		close(natsReceiver)
-		return nil, err
-	}
-
+func NewRoom(roomId string, properties *json.RawMessage, hub *Hub, events AsyncEvents, backend *Backend) (*Room, error) {
 	room := &Room{
 		id:      roomId,
 		hub:     hub,
-		nats:    n,
+		events:  events,
 		backend: backend,
 
 		properties: properties,
 
-		closeChan: make(chan bool, 1),
-		mu:        &sync.RWMutex{},
-		sessions:  make(map[string]Session),
+		closer:   NewCloser(),
+		mu:       &sync.RWMutex{},
+		sessions: make(map[string]Session),
 
 		internalSessions: make(map[Session]bool),
 		virtualSessions:  make(map[*VirtualSession]bool),
@@ -135,11 +118,15 @@ func NewRoom(roomId string, properties *json.RawMessage, hub *Hub, n NatsClient,
 			"room":    roomId,
 		}),
 
-		natsReceiver:        natsReceiver,
-		backendSubscription: backendSubscription,
+		lastRoomRequests: make(map[string]int64),
 
-		lastNatsRoomRequests: make(map[string]int64),
+		transientData: NewTransientData(),
 	}
+
+	if err := events.RegisterBackendRoomListener(roomId, backend, room); err != nil {
+		return nil, err
+	}
+
 	go room.run()
 
 	return room, nil
@@ -186,12 +173,8 @@ func (r *Room) run() {
 loop:
 	for {
 		select {
-		case <-r.closeChan:
+		case <-r.closer.C:
 			break loop
-		case msg := <-r.natsReceiver:
-			if msg != nil {
-				r.processNatsMessage(msg)
-			}
 		case <-ticker.C:
 			r.publishActiveSessions()
 		}
@@ -199,23 +182,11 @@ loop:
 }
 
 func (r *Room) doClose() {
-	select {
-	case r.closeChan <- true:
-	default:
-	}
+	r.closer.Close()
 }
 
 func (r *Room) unsubscribeBackend() {
-	if r.backendSubscription == nil {
-		return
-	}
-
-	go func(subscription NatsSubscription) {
-		if err := subscription.Unsubscribe(); err != nil {
-			log.Printf("Error closing backend subscription for room %s: %s", r.Id(), err)
-		}
-	}(r.backendSubscription)
-	r.backendSubscription = nil
+	r.events.UnregisterBackendRoomListener(r.id, r.backend, r)
 }
 
 func (r *Room) Close() []Session {
@@ -235,33 +206,29 @@ func (r *Room) Close() []Session {
 	return result
 }
 
-func (r *Room) processNatsMessage(message *nats.Msg) {
-	var msg NatsMessage
-	if err := r.nats.Decode(message, &msg); err != nil {
-		log.Printf("Could not decode nats message %+v, %s", message, err)
-		return
-	}
-
-	switch msg.Type {
+func (r *Room) ProcessBackendRoomRequest(message *AsyncMessage) {
+	switch message.Type {
 	case "room":
-		r.processBackendRoomRequest(msg.Room)
+		r.processBackendRoomRequestRoom(message.Room)
+	case "asyncroom":
+		r.processBackendRoomRequestAsyncRoom(message.AsyncRoom)
 	default:
-		log.Printf("Unsupported NATS room request with type %s: %+v", msg.Type, msg)
+		log.Printf("Unsupported backend room request with type %s in %s: %+v", message.Type, r.id, message)
 	}
 }
 
-func (r *Room) processBackendRoomRequest(message *BackendServerRoomRequest) {
+func (r *Room) processBackendRoomRequestRoom(message *BackendServerRoomRequest) {
 	received := message.ReceivedTime
-	if last, found := r.lastNatsRoomRequests[message.Type]; found && last > received {
+	if last, found := r.lastRoomRequests[message.Type]; found && last > received {
 		if msg, err := json.Marshal(message); err == nil {
-			log.Printf("Ignore old NATS backend room request for %s: %s", r.Id(), string(msg))
+			log.Printf("Ignore old backend room request for %s: %s", r.Id(), string(msg))
 		} else {
-			log.Printf("Ignore old NATS backend room request for %s: %+v", r.Id(), message)
+			log.Printf("Ignore old backend room request for %s: %+v", r.Id(), message)
 		}
 		return
 	}
 
-	r.lastNatsRoomRequests[message.Type] = received
+	r.lastRoomRequests[message.Type] = received
 	message.room = r
 	switch message.Type {
 	case "update":
@@ -275,12 +242,26 @@ func (r *Room) processBackendRoomRequest(message *BackendServerRoomRequest) {
 		r.hub.roomParticipants <- message
 	case "message":
 		r.publishRoomMessage(message.Message)
+	case "switchto":
+		r.publishSwitchTo(message.SwitchTo)
 	default:
-		log.Printf("Unsupported NATS backend room request with type %s in %s: %+v", message.Type, r.Id(), message)
+		log.Printf("Unsupported backend room request with type %s in %s: %+v", message.Type, r.Id(), message)
 	}
 }
 
-func (r *Room) AddSession(session Session, sessionData *json.RawMessage) []Session {
+func (r *Room) processBackendRoomRequestAsyncRoom(message *AsyncRoomMessage) {
+	switch message.Type {
+	case "sessionjoined":
+		r.notifySessionJoined(message.SessionId)
+		if message.ClientType == HelloClientTypeInternal {
+			r.publishUsersChangedWithInternal()
+		}
+	default:
+		log.Printf("Unsupported async room request with type %s in %s: %+v", message.Type, r.Id(), message)
+	}
+}
+
+func (r *Room) AddSession(session Session, sessionData *json.RawMessage) {
 	var roomSessionData *RoomSessionData
 	if sessionData != nil && len(*sessionData) > 0 {
 		roomSessionData = &RoomSessionData{}
@@ -293,13 +274,6 @@ func (r *Room) AddSession(session Session, sessionData *json.RawMessage) []Sessi
 	sid := session.PublicId()
 	r.mu.Lock()
 	_, found := r.sessions[sid]
-	// Return list of sessions already in the room.
-	result := make([]Session, 0, len(r.sessions))
-	for _, s := range r.sessions {
-		if s != session {
-			result = append(result, s)
-		}
-	}
 	r.sessions[sid] = session
 	if !found {
 		r.statsRoomSessionsCurrent.With(prometheus.Labels{"clienttype": session.ClientType()}).Inc()
@@ -331,8 +305,111 @@ func (r *Room) AddSession(session Session, sessionData *json.RawMessage) []Sessi
 				r.publishSessionFlagsChanged(session)
 			}
 		}
+		if clientSession, ok := session.(*ClientSession); ok {
+			r.transientData.AddListener(clientSession)
+		}
+	}
+
+	// Trigger notifications that the session joined.
+	if err := r.events.PublishBackendRoomMessage(r.id, r.backend, &AsyncMessage{
+		Type: "asyncroom",
+		AsyncRoom: &AsyncRoomMessage{
+			Type:       "sessionjoined",
+			SessionId:  sid,
+			ClientType: session.ClientType(),
+		},
+	}); err != nil {
+		log.Printf("Error publishing joined event for session %s: %s", sid, err)
+	}
+}
+
+func (r *Room) getOtherSessions(ignoreSessionId string) (Session, []Session) {
+	r.mu.Lock()
+	defer r.mu.Unlock()
+
+	sessions := make([]Session, 0, len(r.sessions))
+	for _, s := range r.sessions {
+		if s.PublicId() == ignoreSessionId {
+			continue
+		}
+
+		sessions = append(sessions, s)
+	}
+
+	return r.sessions[ignoreSessionId], sessions
+}
+
+func (r *Room) notifySessionJoined(sessionId string) {
+	session, sessions := r.getOtherSessions(sessionId)
+	if len(sessions) == 0 {
+		return
+	}
+
+	if session != nil && session.ClientType() != HelloClientTypeClient {
+		session = nil
+	}
+
+	events := make([]*EventServerMessageSessionEntry, 0, len(sessions))
+	for _, s := range sessions {
+		entry := &EventServerMessageSessionEntry{
+			SessionId: s.PublicId(),
+			UserId:    s.UserId(),
+			User:      s.UserData(),
+		}
+		if s, ok := s.(*ClientSession); ok {
+			entry.RoomSessionId = s.RoomSessionId()
+		}
+		events = append(events, entry)
+	}
+
+	msg := &ServerMessage{
+		Type: "event",
+		Event: &EventServerMessage{
+			Target: "room",
+			Type:   "join",
+			Join:   events,
+		},
+	}
+
+	if err := r.events.PublishSessionMessage(sessionId, r.backend, &AsyncMessage{
+		Type:    "message",
+		Message: msg,
+	}); err != nil {
+		log.Printf("Error publishing joined events to session %s: %s", sessionId, err)
+	}
+
+	// Notify about initial flags of virtual sessions.
+	for _, s := range sessions {
+		vsess, ok := s.(*VirtualSession)
+		if !ok {
+			continue
+		}
+
+		flags := vsess.Flags()
+		if flags == 0 {
+			continue
+		}
+
+		msg := &ServerMessage{
+			Type: "event",
+			Event: &EventServerMessage{
+				Target: "participants",
+				Type:   "flags",
+				Flags: &RoomFlagsServerMessage{
+					RoomId:    r.id,
+					SessionId: vsess.PublicId(),
+					Flags:     vsess.Flags(),
+				},
+			},
+		}
+
+		if err := r.events.PublishSessionMessage(sessionId, r.backend, &AsyncMessage{
+			Type:    "message",
+			Message: msg,
+		}); err != nil {
+			log.Printf("Error publishing initial flags to session %s: %s", sessionId, err)
+		}
 	}
-	return result
 }
 
 func (r *Room) HasSession(session Session) bool {
@@ -364,6 +441,9 @@ func (r *Room) RemoveSession(session Session) bool {
 	if virtualSession, ok := session.(*VirtualSession); ok {
 		delete(r.virtualSessions, virtualSession)
 	}
+	if clientSession, ok := session.(*ClientSession); ok {
+		r.transientData.RemoveListener(clientSession)
+	}
 	delete(r.inCallSessions, session)
 	delete(r.roomSessionData, sid)
 	if len(r.sessions) > 0 {
@@ -379,11 +459,16 @@ func (r *Room) RemoveSession(session Session) bool {
 	r.unsubscribeBackend()
 	r.doClose()
 	r.mu.Unlock()
+	// Still need to publish an event so sessions on other servers get notified.
+	r.PublishSessionLeft(session)
 	return false
 }
 
 func (r *Room) publish(message *ServerMessage) error {
-	return r.nats.PublishMessage(GetSubjectForRoomId(r.id, r.backend), message)
+	return r.events.PublishRoomMessage(r.id, r.backend, &AsyncMessage{
+		Type:    "message",
+		Message: message,
+	})
 }
 
 func (r *Room) UpdateProperties(properties *json.RawMessage) {
@@ -445,10 +530,6 @@ func (r *Room) PublishSessionJoined(session Session, sessionData *RoomSessionDat
 	if err := r.publish(message); err != nil {
 		log.Printf("Could not publish session joined message in room %s: %s", r.Id(), err)
 	}
-
-	if session.ClientType() == HelloClientTypeInternal {
-		r.publishUsersChangedWithInternal()
-	}
 }
 
 func (r *Room) PublishSessionLeft(session Session) {
@@ -479,6 +560,7 @@ func (r *Room) PublishSessionLeft(session Session) {
 func (r *Room) addInternalSessions(users []map[string]interface{}) []map[string]interface{} {
 	now := time.Now().Unix()
 	r.mu.Lock()
+	defer r.mu.Unlock()
 	for _, user := range users {
 		sessionid, found := user["sessionId"]
 		if !found || sessionid == "" {
@@ -493,7 +575,7 @@ func (r *Room) addInternalSessions(users []map[string]interface{}) []map[string]
 	}
 	for session := range r.internalSessions {
 		users = append(users, map[string]interface{}{
-			"inCall":    FlagInCall | FlagWithAudio,
+			"inCall":    session.(*ClientSession).GetInCall(),
 			"sessionId": session.PublicId(),
 			"lastPing":  now,
 			"internal":  true,
@@ -501,13 +583,12 @@ func (r *Room) addInternalSessions(users []map[string]interface{}) []map[string]
 	}
 	for session := range r.virtualSessions {
 		users = append(users, map[string]interface{}{
-			"inCall":    FlagInCall | FlagWithPhone,
+			"inCall":    session.GetInCall(),
 			"sessionId": session.PublicId(),
 			"lastPing":  now,
 			"virtual":   true,
 		})
 	}
-	r.mu.Unlock()
 	return users
 }
 
@@ -605,6 +686,93 @@ func (r *Room) PublishUsersInCallChanged(changed []map[string]interface{}, users
 	}
 }
 
+func (r *Room) PublishUsersInCallChangedAll(inCall int) {
+	r.mu.Lock()
+	defer r.mu.Unlock()
+
+	var notify []*ClientSession
+	if inCall&FlagInCall != 0 {
+		// All connected sessions join the call.
+		var joined []string
+		for _, session := range r.sessions {
+			clientSession, ok := session.(*ClientSession)
+			if !ok {
+				continue
+			}
+
+			if session.ClientType() == HelloClientTypeInternal {
+				continue
+			}
+
+			if !r.inCallSessions[session] {
+				r.inCallSessions[session] = true
+				joined = append(joined, session.PublicId())
+			}
+			notify = append(notify, clientSession)
+		}
+
+		if len(joined) == 0 {
+			return
+		}
+
+		log.Printf("Sessions %v joined call %s", joined, r.id)
+	} else if len(r.inCallSessions) > 0 {
+		// Perform actual leaving asynchronously.
+		ch := make(chan *ClientSession, 1)
+		go func() {
+			for {
+				session := <-ch
+				if session == nil {
+					break
+				}
+
+				session.LeaveCall()
+			}
+		}()
+
+		for _, session := range r.sessions {
+			clientSession, ok := session.(*ClientSession)
+			if !ok {
+				continue
+			}
+
+			notify = append(notify, clientSession)
+		}
+
+		for session := range r.inCallSessions {
+			if clientSession, ok := session.(*ClientSession); ok {
+				ch <- clientSession
+			}
+		}
+		close(ch)
+		r.inCallSessions = make(map[Session]bool)
+	} else {
+		// All sessions already left the call, no need to notify.
+		return
+	}
+
+	inCallMsg := json.RawMessage(strconv.FormatInt(int64(inCall), 10))
+
+	message := &ServerMessage{
+		Type: "event",
+		Event: &EventServerMessage{
+			Target: "participants",
+			Type:   "update",
+			Update: &RoomEventServerMessage{
+				RoomId: r.id,
+				InCall: &inCallMsg,
+				All:    true,
+			},
+		},
+	}
+
+	for _, session := range notify {
+		if !session.SendMessage(message) {
+			log.Printf("Could not send incall message from room %s to %s", r.Id(), session.PublicId())
+		}
+	}
+}
+
 func (r *Room) PublishUsersChanged(changed []map[string]interface{}, users []map[string]interface{}) {
 	changed = r.filterPermissions(changed)
 	users = r.filterPermissions(users)
@@ -652,22 +820,59 @@ func (r *Room) NotifySessionResumed(session *ClientSession) {
 	session.SendMessage(message)
 }
 
-func (r *Room) NotifySessionChanged(session Session) {
-	if session.ClientType() != HelloClientTypeVirtual {
+func (r *Room) NotifySessionChanged(session Session, flags SessionChangeFlag) {
+	if flags&SessionChangeFlags != 0 && session.ClientType() == HelloClientTypeVirtual {
 		// Only notify if a virtual session has changed.
-		return
+		if virtual, ok := session.(*VirtualSession); ok {
+			r.publishSessionFlagsChanged(virtual)
+		}
 	}
 
-	virtual, ok := session.(*VirtualSession)
-	if !ok {
-		return
-	}
+	if flags&SessionChangeInCall != 0 {
+		joinLeave := 0
+		if clientSession, ok := session.(*ClientSession); ok {
+			if clientSession.GetInCall()&FlagInCall != 0 {
+				joinLeave = 1
+			} else {
+				joinLeave = 2
+			}
+		} else if virtual, ok := session.(*VirtualSession); ok {
+			if virtual.GetInCall()&FlagInCall != 0 {
+				joinLeave = 1
+			} else {
+				joinLeave = 2
+			}
+		}
 
-	r.publishSessionFlagsChanged(virtual)
+		if joinLeave != 0 {
+			if joinLeave == 1 {
+				r.mu.Lock()
+				if !r.inCallSessions[session] {
+					r.inCallSessions[session] = true
+					log.Printf("Session %s joined call %s", session.PublicId(), r.id)
+				}
+				r.mu.Unlock()
+			} else if joinLeave == 2 {
+				r.mu.Lock()
+				delete(r.inCallSessions, session)
+				r.mu.Unlock()
+				if clientSession, ok := session.(*ClientSession); ok {
+					clientSession.LeaveCall()
+				}
+			}
+
+			// TODO: Check if we could send a smaller update message with only the changed session.
+			r.publishUsersChangedWithInternal()
+		}
+	}
 }
 
 func (r *Room) publishUsersChangedWithInternal() {
 	message := r.getParticipantsUpdateMessage(r.users)
+	if len(message.Event.Update.Users) == 0 {
+		return
+	}
+
 	if err := r.publish(message); err != nil {
 		log.Printf("Could not publish users changed message in room %s: %s", r.Id(), err)
 	}
@@ -746,9 +951,7 @@ func (r *Room) publishActiveSessions() (int, *sync.WaitGroup) {
 			ctx, cancel := context.WithTimeout(context.Background(), r.hub.backendTimeout)
 			defer cancel()
 
-			request := NewBackendClientPingRequest(r.id, entries)
-			var response BackendClientResponse
-			if err := r.hub.backend.PerformJSONRequest(ctx, url, request, &response); err != nil {
+			if err := r.hub.roomPing.SendPings(ctx, r, url, entries); err != nil {
 				log.Printf("Error pinging room %s for active entries %+v: %s", r.id, entries, err)
 			}
 		}(urls[u], e)
@@ -777,6 +980,65 @@ func (r *Room) publishRoomMessage(message *BackendRoomMessageRequest) {
 	}
 }
 
+func (r *Room) publishSwitchTo(message *BackendRoomSwitchToMessageRequest) {
+	var wg sync.WaitGroup
+	if len(message.SessionsList) > 0 {
+		msg := &ServerMessage{
+			Type: "event",
+			Event: &EventServerMessage{
+				Target: "room",
+				Type:   "switchto",
+				SwitchTo: &EventServerMessageSwitchTo{
+					RoomId: message.RoomId,
+				},
+			},
+		}
+
+		for _, sessionId := range message.SessionsList {
+			wg.Add(1)
+			go func(sessionId string) {
+				defer wg.Done()
+
+				if err := r.events.PublishSessionMessage(sessionId, r.backend, &AsyncMessage{
+					Type:    "message",
+					Message: msg,
+				}); err != nil {
+					log.Printf("Error publishing switchto event to session %s: %s", sessionId, err)
+				}
+			}(sessionId)
+		}
+	}
+
+	if len(message.SessionsMap) > 0 {
+		for sessionId, details := range message.SessionsMap {
+			wg.Add(1)
+			go func(sessionId string, details json.RawMessage) {
+				defer wg.Done()
+
+				msg := &ServerMessage{
+					Type: "event",
+					Event: &EventServerMessage{
+						Target: "room",
+						Type:   "switchto",
+						SwitchTo: &EventServerMessageSwitchTo{
+							RoomId:  message.RoomId,
+							Details: details,
+						},
+					},
+				}
+
+				if err := r.events.PublishSessionMessage(sessionId, r.backend, &AsyncMessage{
+					Type:    "message",
+					Message: msg,
+				}); err != nil {
+					log.Printf("Error publishing switchto event to session %s: %s", sessionId, err)
+				}
+			}(sessionId, details)
+		}
+	}
+	wg.Wait()
+}
+
 func (r *Room) notifyInternalRoomDeleted() {
 	msg := &ServerMessage{
 		Type: "event",
@@ -792,3 +1054,11 @@ func (r *Room) notifyInternalRoomDeleted() {
 		s.(*ClientSession).SendMessage(msg)
 	}
 }
+
+func (r *Room) SetTransientData(key string, value interface{}) {
+	r.transientData.Set(key, value)
+}
+
+func (r *Room) RemoveTransientData(key string) {
+	r.transientData.Remove(key)
+}
