@@ -22,6 +22,7 @@
 package signaling
 
 import (
+	"bytes"
 	"context"
 	"crypto/hmac"
 	"crypto/sha256"
@@ -30,10 +31,13 @@ import (
 	"fmt"
 	"net"
 	"net/http/httptest"
+	"reflect"
+	"strconv"
 	"strings"
 	"testing"
 	"time"
 
+	"github.com/golang-jwt/jwt/v4"
 	"github.com/gorilla/websocket"
 )
 
@@ -41,7 +45,7 @@ var (
 	testBackendSecret  = []byte("secret")
 	testInternalSecret = []byte("internal-secret")
 
-	NoMessageReceivedError = fmt.Errorf("No message was received by the server.")
+	ErrNoMessageReceived = fmt.Errorf("no message was received by the server")
 )
 
 type TestBackendClientAuthParams struct {
@@ -77,66 +81,62 @@ func checkUnexpectedClose(err error) error {
 	return nil
 }
 
-func toJsonString(o interface{}) string {
-	if s, err := json.Marshal(o); err != nil {
-		panic(err)
-	} else {
-		return string(s)
-	}
-}
-
 func checkMessageType(message *ServerMessage, expectedType string) error {
 	if message == nil {
-		return NoMessageReceivedError
+		return ErrNoMessageReceived
 	}
 
 	if message.Type != expectedType {
-		return fmt.Errorf("Expected \"%s\" message, got %+v (%s)", expectedType, message, toJsonString(message))
+		return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
 	}
 	switch message.Type {
 	case "hello":
 		if message.Hello == nil {
-			return fmt.Errorf("Expected \"%s\" message, got %+v (%s)", expectedType, message, toJsonString(message))
+			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
 		}
 	case "message":
 		if message.Message == nil {
-			return fmt.Errorf("Expected \"%s\" message, got %+v (%s)", expectedType, message, toJsonString(message))
+			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
 		} else if message.Message.Data == nil || len(*message.Message.Data) == 0 {
 			return fmt.Errorf("Received message without data")
 		}
 	case "room":
 		if message.Room == nil {
-			return fmt.Errorf("Expected \"%s\" message, got %+v (%s)", expectedType, message, toJsonString(message))
+			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
 		}
 	case "event":
 		if message.Event == nil {
-			return fmt.Errorf("Expected \"%s\" message, got %+v (%s)", expectedType, message, toJsonString(message))
+			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
+		}
+	case "transient":
+		if message.TransientData == nil {
+			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
 		}
 	}
 
 	return nil
 }
 
-func checkMessageSender(hub *Hub, message *MessageServerMessage, senderType string, hello *HelloServerMessage) error {
-	if message.Sender.Type != senderType {
-		return fmt.Errorf("Expected sender type %s, got %s", senderType, message.Sender.SessionId)
-	} else if message.Sender.SessionId != hello.SessionId {
+func checkMessageSender(hub *Hub, sender *MessageServerMessageSender, senderType string, hello *HelloServerMessage) error {
+	if sender.Type != senderType {
+		return fmt.Errorf("Expected sender type %s, got %s", senderType, sender.SessionId)
+	} else if sender.SessionId != hello.SessionId {
 		return fmt.Errorf("Expected session id %+v, got %+v",
-			getPubliceSessionIdData(hub, hello.SessionId), getPubliceSessionIdData(hub, message.Sender.SessionId))
-	} else if message.Sender.UserId != hello.UserId {
-		return fmt.Errorf("Expected user id %s, got %s", hello.UserId, message.Sender.UserId)
+			getPubliceSessionIdData(hub, hello.SessionId), getPubliceSessionIdData(hub, sender.SessionId))
+	} else if sender.UserId != hello.UserId {
+		return fmt.Errorf("Expected user id %s, got %s", hello.UserId, sender.UserId)
 	}
 
 	return nil
 }
 
-func checkReceiveClientMessageWithSender(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}, sender **MessageServerMessageSender) error {
+func checkReceiveClientMessageWithSenderAndRecipient(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}, sender **MessageServerMessageSender, recipient **MessageClientMessageRecipient) error {
 	message, err := client.RunUntilMessage(ctx)
 	if err := checkUnexpectedClose(err); err != nil {
 		return err
 	} else if err := checkMessageType(message, "message"); err != nil {
 		return err
-	} else if err := checkMessageSender(client.hub, message.Message, senderType, hello); err != nil {
+	} else if err := checkMessageSender(client.hub, message.Message.Sender, senderType, hello); err != nil {
 		return err
 	} else {
 		if err := json.Unmarshal(*message.Message.Data, payload); err != nil {
@@ -146,11 +146,48 @@ func checkReceiveClientMessageWithSender(ctx context.Context, client *TestClient
 	if sender != nil {
 		*sender = message.Message.Sender
 	}
+	if recipient != nil {
+		*recipient = message.Message.Recipient
+	}
 	return nil
 }
 
+func checkReceiveClientMessageWithSender(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}, sender **MessageServerMessageSender) error {
+	return checkReceiveClientMessageWithSenderAndRecipient(ctx, client, senderType, hello, payload, sender, nil)
+}
+
 func checkReceiveClientMessage(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}) error {
-	return checkReceiveClientMessageWithSender(ctx, client, senderType, hello, payload, nil)
+	return checkReceiveClientMessageWithSenderAndRecipient(ctx, client, senderType, hello, payload, nil, nil)
+}
+
+func checkReceiveClientControlWithSenderAndRecipient(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}, sender **MessageServerMessageSender, recipient **MessageClientMessageRecipient) error {
+	message, err := client.RunUntilMessage(ctx)
+	if err := checkUnexpectedClose(err); err != nil {
+		return err
+	} else if err := checkMessageType(message, "control"); err != nil {
+		return err
+	} else if err := checkMessageSender(client.hub, message.Control.Sender, senderType, hello); err != nil {
+		return err
+	} else {
+		if err := json.Unmarshal(*message.Control.Data, payload); err != nil {
+			return err
+		}
+	}
+	if sender != nil {
+		*sender = message.Control.Sender
+	}
+	if recipient != nil {
+		*recipient = message.Control.Recipient
+	}
+	return nil
+}
+
+func checkReceiveClientControlWithSender(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}, sender **MessageServerMessageSender) error { // nolint
+	return checkReceiveClientControlWithSenderAndRecipient(ctx, client, senderType, hello, payload, sender, nil)
+}
+
+func checkReceiveClientControl(ctx context.Context, client *TestClient, senderType string, hello *HelloServerMessage, payload interface{}) error {
+	return checkReceiveClientControlWithSenderAndRecipient(ctx, client, senderType, hello, payload, nil, nil)
 }
 
 func checkReceiveClientEvent(ctx context.Context, client *TestClient, eventType string, msg **EventServerMessage) error {
@@ -183,9 +220,9 @@ type TestClient struct {
 	publicId string
 }
 
-func NewTestClient(t *testing.T, server *httptest.Server, hub *Hub) *TestClient {
+func NewTestClientContext(ctx context.Context, t *testing.T, server *httptest.Server, hub *Hub) *TestClient {
 	// Reference "hub" to prevent compiler error.
-	conn, _, err := websocket.DefaultDialer.Dial(getWebsocketUrl(server.URL), nil)
+	conn, _, err := websocket.DefaultDialer.DialContext(ctx, getWebsocketUrl(server.URL), nil)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -221,6 +258,22 @@ func NewTestClient(t *testing.T, server *httptest.Server, hub *Hub) *TestClient
 	}
 }
 
+func NewTestClient(t *testing.T, server *httptest.Server, hub *Hub) *TestClient {
+	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
+	defer cancel()
+
+	client := NewTestClientContext(ctx, t, server, hub)
+	msg, err := client.RunUntilMessage(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if msg.Type != "welcome" {
+		t.Errorf("Expected welcome message, got %+v", msg)
+	}
+	return client
+}
+
 func (c *TestClient) CloseWithBye() {
 	c.SendBye() // nolint
 	c.Close()
@@ -269,6 +322,7 @@ func (c *TestClient) WaitForClientRemoved(ctx context.Context) error {
 		c.hub.mu.Unlock()
 		select {
 		case <-ctx.Done():
+			c.hub.mu.Lock()
 			return ctx.Err()
 		default:
 			time.Sleep(time.Millisecond)
@@ -295,6 +349,7 @@ func (c *TestClient) WaitForSessionRemoved(ctx context.Context, sessionId string
 		c.hub.mu.Unlock()
 		select {
 		case <-ctx.Done():
+			c.hub.mu.Lock()
 			return ctx.Err()
 		default:
 			time.Sleep(time.Millisecond)
@@ -305,11 +360,14 @@ func (c *TestClient) WaitForSessionRemoved(ctx context.Context, sessionId string
 }
 
 func (c *TestClient) WriteJSON(data interface{}) error {
-	if msg, ok := data.(*ClientMessage); ok {
-		if err := msg.CheckValid(); err != nil {
-			return err
+	if !strings.Contains(c.t.Name(), "HelloUnsupportedVersion") {
+		if msg, ok := data.(*ClientMessage); ok {
+			if err := msg.CheckValid(); err != nil {
+				return err
+			}
 		}
 	}
+
 	return c.conn.WriteJSON(data)
 }
 
@@ -320,10 +378,63 @@ func (c *TestClient) EnsuerWriteJSON(data interface{}) {
 }
 
 func (c *TestClient) SendHello(userid string) error {
+	return c.SendHelloV1(userid)
+}
+
+func (c *TestClient) SendHelloV1(userid string) error {
 	params := TestBackendClientAuthParams{
 		UserId: userid,
 	}
-	return c.SendHelloParams(c.server.URL, "", params)
+	return c.SendHelloParams(c.server.URL, HelloVersionV1, "", nil, params)
+}
+
+func (c *TestClient) SendHelloV2(userid string) error {
+	now := time.Now()
+	return c.SendHelloV2WithTimes(userid, now, now.Add(time.Minute))
+}
+
+func (c *TestClient) SendHelloV2WithTimes(userid string, issuedAt time.Time, expiresAt time.Time) error {
+	userdata := map[string]string{
+		"displayname": "Displayname " + userid,
+	}
+
+	data, err := json.Marshal(userdata)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+
+	claims := &HelloV2TokenClaims{
+		RegisteredClaims: jwt.RegisteredClaims{
+			Issuer:  c.server.URL,
+			Subject: userid,
+		},
+		UserData: (*json.RawMessage)(&data),
+	}
+	if !issuedAt.IsZero() {
+		claims.IssuedAt = jwt.NewNumericDate(issuedAt)
+	}
+	if !expiresAt.IsZero() {
+		claims.ExpiresAt = jwt.NewNumericDate(expiresAt)
+	}
+
+	var token *jwt.Token
+	if strings.Contains(c.t.Name(), "ECDSA") {
+		token = jwt.NewWithClaims(jwt.SigningMethodES256, claims)
+	} else if strings.Contains(c.t.Name(), "Ed25519") {
+		token = jwt.NewWithClaims(jwt.SigningMethodEdDSA, claims)
+	} else {
+		token = jwt.NewWithClaims(jwt.SigningMethodRS256, claims)
+	}
+	private := getPrivateAuthToken(c.t)
+	tokenString, err := token.SignedString(private)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+
+	params := HelloV2AuthParams{
+		Token: tokenString,
+	}
+	return c.SendHelloParams(c.server.URL, HelloVersionV2, "", nil, params)
 }
 
 func (c *TestClient) SendHelloResume(resumeId string) error {
@@ -331,7 +442,7 @@ func (c *TestClient) SendHelloResume(resumeId string) error {
 		Id:   "1234",
 		Type: "hello",
 		Hello: &HelloClientMessage{
-			Version:  HelloVersion,
+			Version:  HelloVersionV1,
 			ResumeId: resumeId,
 		},
 	}
@@ -342,10 +453,14 @@ func (c *TestClient) SendHelloClient(userid string) error {
 	params := TestBackendClientAuthParams{
 		UserId: userid,
 	}
-	return c.SendHelloParams(c.server.URL, "client", params)
+	return c.SendHelloParams(c.server.URL, HelloVersionV1, "client", nil, params)
 }
 
 func (c *TestClient) SendHelloInternal() error {
+	return c.SendHelloInternalWithFeatures(nil)
+}
+
+func (c *TestClient) SendHelloInternalWithFeatures(features []string) error {
 	random := newRandomString(48)
 	mac := hmac.New(sha256.New, testInternalSecret)
 	mac.Write([]byte(random)) // nolint
@@ -357,10 +472,10 @@ func (c *TestClient) SendHelloInternal() error {
 		Token:   token,
 		Backend: backend,
 	}
-	return c.SendHelloParams("", "internal", params)
+	return c.SendHelloParams("", HelloVersionV1, "internal", features, params)
 }
 
-func (c *TestClient) SendHelloParams(url string, clientType string, params interface{}) error {
+func (c *TestClient) SendHelloParams(url string, version string, clientType string, features []string, params interface{}) error {
 	data, err := json.Marshal(params)
 	if err != nil {
 		c.t.Fatal(err)
@@ -370,7 +485,8 @@ func (c *TestClient) SendHelloParams(url string, clientType string, params inter
 		Id:   "1234",
 		Type: "hello",
 		Hello: &HelloClientMessage{
-			Version: HelloVersion,
+			Version:  version,
+			Features: features,
 			Auth: HelloClientMessageAuth{
 				Type:   clientType,
 				Url:    url,
@@ -407,6 +523,91 @@ func (c *TestClient) SendMessage(recipient MessageClientMessageRecipient, data i
 	return c.WriteJSON(message)
 }
 
+func (c *TestClient) SendControl(recipient MessageClientMessageRecipient, data interface{}) error {
+	payload, err := json.Marshal(data)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+
+	message := &ClientMessage{
+		Id:   "abcd",
+		Type: "control",
+		Control: &ControlClientMessage{
+			MessageClientMessage: MessageClientMessage{
+				Recipient: recipient,
+				Data:      (*json.RawMessage)(&payload),
+			},
+		},
+	}
+	return c.WriteJSON(message)
+}
+
+func (c *TestClient) SendInternalAddSession(msg *AddSessionInternalClientMessage) error {
+	message := &ClientMessage{
+		Id:   "abcd",
+		Type: "internal",
+		Internal: &InternalClientMessage{
+			Type:       "addsession",
+			AddSession: msg,
+		},
+	}
+	return c.WriteJSON(message)
+}
+
+func (c *TestClient) SendInternalUpdateSession(msg *UpdateSessionInternalClientMessage) error {
+	message := &ClientMessage{
+		Id:   "abcd",
+		Type: "internal",
+		Internal: &InternalClientMessage{
+			Type:          "updatesession",
+			UpdateSession: msg,
+		},
+	}
+	return c.WriteJSON(message)
+}
+
+func (c *TestClient) SendInternalRemoveSession(msg *RemoveSessionInternalClientMessage) error {
+	message := &ClientMessage{
+		Id:   "abcd",
+		Type: "internal",
+		Internal: &InternalClientMessage{
+			Type:          "removesession",
+			RemoveSession: msg,
+		},
+	}
+	return c.WriteJSON(message)
+}
+
+func (c *TestClient) SetTransientData(key string, value interface{}) error {
+	payload, err := json.Marshal(value)
+	if err != nil {
+		c.t.Fatal(err)
+	}
+
+	message := &ClientMessage{
+		Id:   "efgh",
+		Type: "transient",
+		TransientData: &TransientDataClientMessage{
+			Type:  "set",
+			Key:   key,
+			Value: (*json.RawMessage)(&payload),
+		},
+	}
+	return c.WriteJSON(message)
+}
+
+func (c *TestClient) RemoveTransientData(key string) error {
+	message := &ClientMessage{
+		Id:   "ijkl",
+		Type: "transient",
+		TransientData: &TransientDataClientMessage{
+			Type: "remove",
+			Key:  key,
+		},
+	}
+	return c.WriteJSON(message)
+}
+
 func (c *TestClient) DrainMessages(ctx context.Context) error {
 	select {
 	case err := <-c.readErrorChan:
@@ -546,40 +747,70 @@ func (c *TestClient) checkSingleMessageJoined(message *ServerMessage) error {
 func (c *TestClient) checkMessageJoinedSession(message *ServerMessage, sessionId string, userId string) error {
 	if err := c.checkSingleMessageJoined(message); err != nil {
 		return err
-	} else {
-		evt := message.Event.Join[0]
-		if sessionId != "" && evt.SessionId != sessionId {
-			return fmt.Errorf("Expected join session id %+v, got %+v",
-				getPubliceSessionIdData(c.hub, sessionId), getPubliceSessionIdData(c.hub, evt.SessionId))
-		}
-		if evt.UserId != userId {
-			return fmt.Errorf("Expected join user id %s, got %+v", userId, evt)
-		}
+	}
+
+	evt := message.Event.Join[0]
+	if sessionId != "" && evt.SessionId != sessionId {
+		return fmt.Errorf("Expected join session id %+v, got %+v",
+			getPubliceSessionIdData(c.hub, sessionId), getPubliceSessionIdData(c.hub, evt.SessionId))
+	}
+	if evt.UserId != userId {
+		return fmt.Errorf("Expected join user id %s, got %+v", userId, evt)
 	}
 	return nil
 }
 
-func (c *TestClient) RunUntilJoined(ctx context.Context, hello ...*HelloServerMessage) error {
-	for len(hello) > 0 {
-		if message, err := c.RunUntilMessage(ctx); err != nil {
-			return err
-		} else {
-			if err := c.checkSingleMessageJoined(message); err != nil {
-				return err
-			}
+func (c *TestClient) RunUntilJoinedAndReturn(ctx context.Context, hello ...*HelloServerMessage) ([]*EventServerMessageSessionEntry, []*ServerMessage, error) {
+	received := make([]*EventServerMessageSessionEntry, len(hello))
+	var ignored []*ServerMessage
+	hellos := make(map[*HelloServerMessage]int, len(hello))
+	for idx, h := range hello {
+		hellos[h] = idx
+	}
+	for len(hellos) > 0 {
+		message, err := c.RunUntilMessage(ctx)
+		if err != nil {
+			return nil, nil, fmt.Errorf("got error while waiting for %+v: %w", hellos, err)
+		}
+
+		if err := checkMessageType(message, "event"); err != nil {
+			ignored = append(ignored, message)
+			continue
+		} else if message.Event.Target != "room" || message.Event.Type != "join" {
+			ignored = append(ignored, message)
+			continue
+		}
+
+		for len(message.Event.Join) > 0 {
 			found := false
-			for idx, h := range hello {
-				if err := c.checkMessageJoined(message, h); err == nil {
-					hello = append(hello[:idx], hello[idx+1:]...)
-					found = true
-					break
+		loop:
+			for h, idx := range hellos {
+				for idx2, evt := range message.Event.Join {
+					if evt.SessionId == h.SessionId && evt.UserId == h.UserId {
+						received[idx] = evt
+						delete(hellos, h)
+						message.Event.Join = append(message.Event.Join[:idx2], message.Event.Join[idx2+1:]...)
+						found = true
+						break loop
+					}
 				}
 			}
 			if !found {
-				return fmt.Errorf("expected one of the passed hello sessions, got %+v", message.Event.Join[0])
+				return nil, nil, fmt.Errorf("expected one of the passed hello sessions, got %+v", message.Event.Join[0])
 			}
 		}
 	}
+	return received, ignored, nil
+}
+
+func (c *TestClient) RunUntilJoined(ctx context.Context, hello ...*HelloServerMessage) error {
+	_, unexpected, err := c.RunUntilJoinedAndReturn(ctx, hello...)
+	if err != nil {
+		return err
+	}
+	if len(unexpected) > 0 {
+		return fmt.Errorf("Received unexpected messages: %+v", unexpected)
+	}
 	return nil
 }
 
@@ -604,11 +835,12 @@ func (c *TestClient) checkMessageRoomLeaveSession(message *ServerMessage, sessio
 }
 
 func (c *TestClient) RunUntilLeft(ctx context.Context, hello *HelloServerMessage) error {
-	if message, err := c.RunUntilMessage(ctx); err != nil {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
 		return err
-	} else {
-		return c.checkMessageRoomLeave(message, hello)
 	}
+
+	return c.checkMessageRoomLeave(message, hello)
 }
 
 func checkMessageRoomlistUpdate(message *ServerMessage) (*RoomEventServerMessage, error) {
@@ -624,11 +856,12 @@ func checkMessageRoomlistUpdate(message *ServerMessage) (*RoomEventServerMessage
 }
 
 func (c *TestClient) RunUntilRoomlistUpdate(ctx context.Context) (*RoomEventServerMessage, error) {
-	if message, err := c.RunUntilMessage(ctx); err != nil {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
 		return nil, err
-	} else {
-		return checkMessageRoomlistUpdate(message)
 	}
+
+	return checkMessageRoomlistUpdate(message)
 }
 
 func checkMessageRoomlistDisinvite(message *ServerMessage) (*RoomDisinviteEventServerMessage, error) {
@@ -644,20 +877,21 @@ func checkMessageRoomlistDisinvite(message *ServerMessage) (*RoomDisinviteEventS
 }
 
 func (c *TestClient) RunUntilRoomlistDisinvite(ctx context.Context) (*RoomDisinviteEventServerMessage, error) {
-	if message, err := c.RunUntilMessage(ctx); err != nil {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
 		return nil, err
-	} else {
-		return checkMessageRoomlistDisinvite(message)
 	}
+
+	return checkMessageRoomlistDisinvite(message)
 }
 
 func checkMessageParticipantsInCall(message *ServerMessage) (*RoomEventServerMessage, error) {
 	if err := checkMessageType(message, "event"); err != nil {
 		return nil, err
 	} else if message.Event.Target != "participants" {
-		return nil, fmt.Errorf("Expected event target room, got %+v", message.Event)
+		return nil, fmt.Errorf("Expected event target participants, got %+v", message.Event)
 	} else if message.Event.Type != "update" || message.Event.Update == nil {
-		return nil, fmt.Errorf("Expected event type incall, got %+v", message.Event)
+		return nil, fmt.Errorf("Expected event type update, got %+v", message.Event)
 	}
 
 	return message.Event.Update, nil
@@ -688,11 +922,12 @@ func checkMessageRoomMessage(message *ServerMessage) (*RoomEventMessage, error)
 }
 
 func (c *TestClient) RunUntilRoomMessage(ctx context.Context) (*RoomEventMessage, error) {
-	if message, err := c.RunUntilMessage(ctx); err != nil {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
 		return nil, err
-	} else {
-		return checkMessageRoomMessage(message)
 	}
+
+	return checkMessageRoomMessage(message)
 }
 
 func checkMessageError(message *ServerMessage, msgid string) error {
@@ -705,6 +940,37 @@ func checkMessageError(message *ServerMessage, msgid string) error {
 	return nil
 }
 
+func (c *TestClient) RunUntilOffer(ctx context.Context, offer string) error {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
+		return err
+	}
+	if err := checkUnexpectedClose(err); err != nil {
+		return err
+	} else if err := checkMessageType(message, "message"); err != nil {
+		return err
+	}
+
+	var data map[string]interface{}
+	if err := json.Unmarshal(*message.Message.Data, &data); err != nil {
+		return err
+	}
+
+	if data["type"].(string) != "offer" {
+		return fmt.Errorf("expected data type offer, got %+v", data)
+	}
+
+	payload := data["payload"].(map[string]interface{})
+	if payload["type"].(string) != "offer" {
+		return fmt.Errorf("expected payload type offer, got %+v", payload)
+	}
+	if payload["sdp"].(string) != offer {
+		return fmt.Errorf("expected payload answer %s, got %+v", offer, payload)
+	}
+
+	return nil
+}
+
 func (c *TestClient) RunUntilAnswer(ctx context.Context, answer string) error {
 	message, err := c.RunUntilMessage(ctx)
 	if err != nil {
@@ -735,3 +1001,91 @@ func (c *TestClient) RunUntilAnswer(ctx context.Context, answer string) error {
 
 	return nil
 }
+
+func checkMessageTransientSet(message *ServerMessage, key string, value interface{}, oldValue interface{}) error {
+	if err := checkMessageType(message, "transient"); err != nil {
+		return err
+	} else if message.TransientData.Type != "set" {
+		return fmt.Errorf("Expected transient set, got %+v", message.TransientData)
+	} else if message.TransientData.Key != key {
+		return fmt.Errorf("Expected transient set key %s, got %+v", key, message.TransientData)
+	} else if !reflect.DeepEqual(message.TransientData.Value, value) {
+		return fmt.Errorf("Expected transient set value %+v, got %+v", value, message.TransientData.Value)
+	} else if !reflect.DeepEqual(message.TransientData.OldValue, oldValue) {
+		return fmt.Errorf("Expected transient set old value %+v, got %+v", oldValue, message.TransientData.OldValue)
+	}
+
+	return nil
+}
+
+func checkMessageTransientRemove(message *ServerMessage, key string, oldValue interface{}) error {
+	if err := checkMessageType(message, "transient"); err != nil {
+		return err
+	} else if message.TransientData.Type != "remove" {
+		return fmt.Errorf("Expected transient remove, got %+v", message.TransientData)
+	} else if message.TransientData.Key != key {
+		return fmt.Errorf("Expected transient remove key %s, got %+v", key, message.TransientData)
+	} else if !reflect.DeepEqual(message.TransientData.OldValue, oldValue) {
+		return fmt.Errorf("Expected transient remove old value %+v, got %+v", oldValue, message.TransientData.OldValue)
+	}
+
+	return nil
+}
+
+func checkMessageTransientInitial(message *ServerMessage, data map[string]interface{}) error {
+	if err := checkMessageType(message, "transient"); err != nil {
+		return err
+	} else if message.TransientData.Type != "initial" {
+		return fmt.Errorf("Expected transient initial, got %+v", message.TransientData)
+	} else if !reflect.DeepEqual(message.TransientData.Data, data) {
+		return fmt.Errorf("Expected transient initial data %+v, got %+v", data, message.TransientData.Data)
+	}
+
+	return nil
+}
+
+func checkMessageInCallAll(message *ServerMessage, roomId string, inCall int) error {
+	if err := checkMessageType(message, "event"); err != nil {
+		return err
+	} else if message.Event.Type != "update" {
+		return fmt.Errorf("Expected update event, got %+v", message.Event)
+	} else if message.Event.Target != "participants" {
+		return fmt.Errorf("Expected participants update event, got %+v", message.Event)
+	} else if message.Event.Update.RoomId != roomId {
+		return fmt.Errorf("Expected participants update event for room %s, got %+v", roomId, message.Event.Update)
+	} else if !message.Event.Update.All {
+		return fmt.Errorf("Expected participants update event for all, got %+v", message.Event.Update)
+	} else if !bytes.Equal(*message.Event.Update.InCall, []byte(strconv.FormatInt(int64(inCall), 10))) {
+		return fmt.Errorf("Expected incall flags %d, got %+v", inCall, message.Event.Update)
+	}
+	return nil
+}
+
+func checkMessageSwitchTo(message *ServerMessage, roomId string, details json.RawMessage) (*EventServerMessageSwitchTo, error) {
+	if err := checkMessageType(message, "event"); err != nil {
+		return nil, err
+	} else if message.Event.Type != "switchto" {
+		return nil, fmt.Errorf("Expected switchto event, got %+v", message.Event)
+	} else if message.Event.Target != "room" {
+		return nil, fmt.Errorf("Expected room switchto event, got %+v", message.Event)
+	} else if message.Event.SwitchTo.RoomId != roomId {
+		return nil, fmt.Errorf("Expected room switchto event for room %s, got %+v", roomId, message.Event)
+	}
+	if details != nil {
+		if message.Event.SwitchTo.Details == nil || !bytes.Equal(details, message.Event.SwitchTo.Details) {
+			return nil, fmt.Errorf("Expected details %s, got %+v", string(details), message.Event)
+		}
+	} else if message.Event.SwitchTo.Details != nil {
+		return nil, fmt.Errorf("Expected no details, got %+v", message.Event)
+	}
+	return message.Event.SwitchTo, nil
+}
+
+func (c *TestClient) RunUntilSwitchTo(ctx context.Context, roomId string, details json.RawMessage) (*EventServerMessageSwitchTo, error) {
+	message, err := c.RunUntilMessage(ctx)
+	if err != nil {
+		return nil, err
+	}
+
+	return checkMessageSwitchTo(message, roomId, details)
+}
